from typing import Iterator
from unittest import mock

import pytest
from requests import Response
from truss.remote.truss_remote import TrussService

TEST_SERVICE_URL = "http://test.com"


class TrussTestService(TrussService):
    def __init__(self, _service_url: str, is_draft: bool, **kwargs):
        super().__init__(_service_url, is_draft, **kwargs)

    def authenticate(self):
        return {}

    def is_live(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    def is_ready(self) -> bool:
        response = self._send_request(self._service_url, "GET")
        if response.status_code == 200:
            return True
        return False

    @property
    def logs_url(self) -> str:
        raise NotImplementedError()

    @property
    def predict_url(self) -> str:
        return f"{self._service_url}/v1/models/model:predict"

    def poll_deployment_status(self, sleep_secs: int = 1) -> Iterator[str]:
        for status in ["DEPLOYING", "ACTIVE"]:
            yield status


def mock_successful_response():
    response = Response()
    response.status_code = 200
    response.json = mock.Mock(return_value={"data": {"status": "success"}})
    return response


def mock_unsuccessful_response():
    response = Response()
    response.status_code = 404
    return response


@mock.patch("requests.request", return_value=mock_successful_response())
def test_is_live(mock_request):
    service = TrussTestService(TEST_SERVICE_URL, True)
    assert service.is_live()


@mock.patch("requests.request", return_value=mock_unsuccessful_response())
def test_is_not_live(mock_request):
    service = TrussTestService(TEST_SERVICE_URL, True)
    assert service.is_live() is False


@mock.patch("requests.request", return_value=mock_successful_response())
def test_is_ready(mock_request):
    service = TrussTestService(TEST_SERVICE_URL, True)
    assert service.is_ready()


@mock.patch("requests.request", return_value=mock_unsuccessful_response())
def test_is_not_ready(mock_request):
    service = TrussTestService(TEST_SERVICE_URL, True)
    assert service.is_ready() is False


@mock.patch("requests.request", return_value=mock_successful_response())
def test_predict(mock_request):
    service = TrussTestService(TEST_SERVICE_URL, True)
    response = service.predict({"model_input": "test"})
    assert response.status_code == 200


@mock.patch("requests.request", return_value=mock_successful_response())
def test_predict_no_data(mock_request):
    service = TrussTestService(TEST_SERVICE_URL, True)
    with pytest.raises(ValueError):
        service._send_request(TEST_SERVICE_URL, "POST")



import copy
import inspect
from typing import Dict

from truss.base.truss_spec import TrussSpec
from truss.local.local_config_handler import LocalConfigHandler


def prepare_secrets(spec: TrussSpec) -> Dict[str, str]:
    secrets = copy.deepcopy(spec.secrets)
    local_secerts = LocalConfigHandler.get_config().secrets
    for secret_name in secrets:
        if secret_name in local_secerts:
            secrets[secret_name] = local_secerts[secret_name]
    return secrets


def signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False



import logging
import os
import shutil
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, List, Tuple

import pydantic
import requests
import yaml

try:
    from shared.util import BLOB_DOWNLOAD_TIMEOUT_SECS
except ModuleNotFoundError:
    from truss.templates.shared.util import BLOB_DOWNLOAD_TIMEOUT_SECS

LAZY_DATA_RESOLVER_PATH = Path("/bptr/bptr-manifest")
NUM_WORKERS = 4
CACHE_DIR = Path("/cache/org/artifacts")
BASETEN_FS_ENABLED_ENV_VAR = "BASETEN_FS_ENABLED"

logger = logging.getLogger(__name__)


class Resolution(pydantic.BaseModel):
    url: str
    expiration_timestamp: int


class BasetenPointer(pydantic.BaseModel):
    """Specification for lazy data resolution for download of large files, similar to Git LFS pointers"""

    resolution: Resolution
    uid: str
    file_name: str
    hashtype: str
    hash: str
    size: int


class BasetenPointerManifest(pydantic.BaseModel):
    pointers: List[BasetenPointer]


class LazyDataResolver:
    def __init__(self, data_dir: Path):
        self._data_dir: Path = data_dir
        self._bptr_resolution: Dict[str, Tuple[str, str]] = _read_bptr_resolution()
        self._resolution_done = False
        self._uses_b10_cache = (
            os.environ.get(BASETEN_FS_ENABLED_ENV_VAR, "False") == "True"
        )

    def cached_download_from_url_using_requests(
        self, URL: str, hash: str, file_name: str
    ):
        """Download object from URL, attempt to write to cache and symlink to data directory if applicable, data directory otherwise.
        In case of failure, write to data directory
        """
        if self._uses_b10_cache:
            file_path = CACHE_DIR / hash
            if file_path.exists():
                os.symlink(file_path, self._data_dir / file_name)
                return

        # Streaming download to keep memory usage low
        resp = requests.get(
            URL,
            allow_redirects=True,
            stream=True,
            timeout=BLOB_DOWNLOAD_TIMEOUT_SECS,
        )
        resp.raise_for_status()

        if self._uses_b10_cache:
            try:
                file_path.parent.mkdir(parents=True, exist_ok=True)
                with file_path.open("wb") as file:
                    shutil.copyfileobj(resp.raw, file)
                # symlink to data directory
                os.symlink(file_path, self._data_dir / file_name)
                return
            except OSError:
                logger.debug(
                    "Failed to save artifact to cache dir, saving to data dir instead"
                )
                # Cache likely has no space left on device, break to download to data dir as fallback
                pass

        file_path = self._data_dir / file_name
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with file_path.open("wb") as file:
            shutil.copyfileobj(resp.raw, file)

    def fetch(self):
        if self._resolution_done:
            return

        with ThreadPoolExecutor(NUM_WORKERS) as executor:
            futures = {}
            for file_name, (resolved_url, hash) in self._bptr_resolution.items():
                futures[file_name] = executor.submit(
                    self.cached_download_from_url_using_requests,
                    resolved_url,
                    hash,
                    file_name,
                )
            for file_name, future in futures.items():
                if not future:
                    raise RuntimeError(f"Download failure for file {file_name}")
        self._resolution_done = True


def _read_bptr_resolution() -> Dict[str, Tuple[str, str]]:
    if not LAZY_DATA_RESOLVER_PATH.is_file():
        return {}
    bptr_manifest = BasetenPointerManifest(
        **yaml.safe_load(LAZY_DATA_RESOLVER_PATH.read_text())
    )
    resolution_map = {}
    for bptr in bptr_manifest.pointers:
        if bptr.resolution.expiration_timestamp < int(
            datetime.now(timezone.utc).timestamp()
        ):
            raise RuntimeError("Baseten pointer lazy data resolution has expired")
        resolution_map[bptr.file_name] = bptr.resolution.url, bptr.hash
    return resolution_map



from pathlib import Path
from typing import List

from truss.truss_handle.patch.dir_signature import directory_content_signature
from truss.truss_handle.truss_gatherer import gather


def test_gather(custom_model_with_external_package):
    gathered_truss_path = gather(custom_model_with_external_package)
    subdir = gathered_truss_path / "packages" / "subdir"
    sub_module = gathered_truss_path / "packages" / "subdir" / "sub_module.py"
    ext_pkg_top_module = gathered_truss_path / "packages" / "top_module.py"
    ext_pkg_top_module2 = gathered_truss_path / "packages" / "top_module2.py"
    assert subdir.exists()
    assert ext_pkg_top_module.exists()
    assert ext_pkg_top_module2.exists()
    assert sub_module.exists()
    sub_module.unlink()
    subdir.rmdir()
    ext_pkg_top_module.unlink()
    ext_pkg_top_module2.unlink()

    assert _same_dir_content(
        custom_model_with_external_package,
        gathered_truss_path,
        ["config.yaml"],
    )


def _same_dir_content(dir1: Path, dir2: Path, ignore_paths: List[str]) -> bool:
    sig1 = directory_content_signature(dir1)
    for path in ignore_paths:
        del sig1[path]
    sig2 = directory_content_signature(dir2)
    for path in ignore_paths:
        del sig2[path]
    return sig1 == sig2



import asyncio
import builtins
import contextlib
import json
import logging
import sys
import textwrap
import threading
import traceback
from typing import (
    Dict,
    Iterator,
    Mapping,
    NoReturn,
    Type,
    TypeVar,
)

import aiohttp
import fastapi
import httpx
import pydantic
from truss.templates.shared import dynamic_config_resolver

from truss_chains import definitions

T = TypeVar("T")


def populate_chainlet_service_predict_urls(
    chainlet_to_service: Mapping[str, definitions.ServiceDescriptor],
) -> Mapping[str, definitions.DeployedServiceDescriptor]:
    chainlet_to_deployed_service: Dict[str, definitions.DeployedServiceDescriptor] = {}

    dynamic_chainlet_config_str = dynamic_config_resolver.get_dynamic_config_value_sync(
        definitions.DYNAMIC_CHAINLET_CONFIG_KEY
    )

    if not dynamic_chainlet_config_str:
        raise definitions.MissingDependencyError(
            f"No '{definitions.DYNAMIC_CHAINLET_CONFIG_KEY}' "
            "found. Cannot override Chainlet configs."
        )

    dynamic_chainlet_config = json.loads(dynamic_chainlet_config_str)

    for (
        chainlet_name,
        service_descriptor,
    ) in chainlet_to_service.items():
        display_name = service_descriptor.display_name

        # NOTE: The Chainlet `display_name` in the Truss CLI
        # corresponds to Chainlet `name` in the backend. As
        # the dynamic Chainlet config is keyed on the backend
        # Chainlet name, we have to look up config values by
        # using the `display_name` in the service descriptor.
        if display_name not in dynamic_chainlet_config:
            raise definitions.MissingDependencyError(
                f"Chainlet '{display_name}' not found in "
                f"'{definitions.DYNAMIC_CHAINLET_CONFIG_KEY}'. "
                f"Dynamic Chainlet config keys: {list(dynamic_chainlet_config)}."
            )

        chainlet_to_deployed_service[chainlet_name] = (
            definitions.DeployedServiceDescriptor(
                display_name=display_name,
                name=service_descriptor.name,
                options=service_descriptor.options,
                predict_url=dynamic_chainlet_config[display_name]["predict_url"],
            )
        )

    return chainlet_to_deployed_service


class AsyncSafeCounter:
    def __init__(self, initial: int = 0) -> None:
        self._counter = initial
        self._lock = asyncio.Lock()

    async def increment(self) -> int:
        async with self._lock:
            self._counter += 1
            return self._counter

    async def decrement(self) -> int:
        async with self._lock:
            self._counter -= 1
            return self._counter

    async def __aenter__(self) -> int:
        return await self.increment()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.decrement()


class ThreadSafeCounter:
    def __init__(self, initial: int = 0) -> None:
        self._counter = initial
        self._lock = threading.Lock()

    def increment(self) -> int:
        with self._lock:
            self._counter += 1
            return self._counter

    def decrement(self) -> int:
        with self._lock:
            self._counter -= 1
            return self._counter

    def __enter__(self) -> int:
        return self.increment()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.decrement()


def pydantic_set_field_dict(obj: pydantic.BaseModel) -> dict[str, pydantic.BaseModel]:
    """Like `BaseModel.model_dump(exclude_unset=True), but only top-level.

    This is used to get kwargs for invoking a function, while dropping fields for which
    there is no value explicitly set in the pydantic model. A field is considered unset
    if the key was not present in the incoming JSON request (from which the model was
    parsed/initialized) and the pydantic model has a default value, such as `None`.

    By dropping these unset fields, the default values from the function definition
    will be used instead. This behavior ensures correct handling of arguments where
    the function has a default, such as in the case of `run_remote`. If the model has
    an optional field defaulting to `None`, this approach differentiates between
    the user explicitly passing a value of `None` and the field being unset in the
    request.

    """
    return {name: getattr(obj, name) for name in obj.model_fields_set}


# Error Propagation Utils. #############################################################


def _handle_exception(exception: Exception) -> NoReturn:
    """Raises `HTTPException` with `RemoteErrorDetail`."""
    if hasattr(exception, "__module__"):
        exception_module_name = exception.__module__
    else:
        exception_module_name = None

    error_stack = traceback.extract_tb(exception.__traceback__)
    # Filter everything before (model.py) and after (stubs, error handling) so that only
    # user-defined code remains. See test_e2e.py::test_chain for expected results.
    model_predict_index = 0
    first_stub_index = len(error_stack)
    for i, frame in enumerate(error_stack):
        if frame.filename.endswith("model/model.py") and frame.name == "predict":
            model_predict_index = i + 1
        if frame.filename.endswith("remote_chainlet/stub.py") and frame.name.startswith(
            "predict"  # predict sycnc|async|stream.
        ):
            first_stub_index = i - 1
            break

    final_tb = error_stack[model_predict_index:first_stub_index]
    stack = [definitions.StackFrame.from_frame_summary(frame) for frame in final_tb]
    error = definitions.RemoteErrorDetail(
        exception_cls_name=exception.__class__.__name__,
        exception_module_name=exception_module_name,
        exception_message=str(exception),
        user_stack_trace=list(stack),
    )
    raise fastapi.HTTPException(
        status_code=500, detail=error.model_dump()
    ) from exception


@contextlib.contextmanager
def exception_to_http_error() -> Iterator[None]:
    try:
        yield
    except Exception as e:
        _handle_exception(e)


def _resolve_exception_class(
    error: definitions.RemoteErrorDetail,
) -> Type[Exception]:
    """Tries to find the exception class in builtins or imported libs,
    falls back to `definitions.GenericRemoteError` if not found."""
    exception_cls = None
    if error.exception_module_name is None:
        exception_cls = getattr(builtins, error.exception_cls_name, None)
    else:
        if mod := sys.modules.get(error.exception_module_name):
            exception_cls = getattr(mod, error.exception_cls_name, None)

    if exception_cls is None:
        logging.warning(
            f"Could not resolve exception with name `{error.exception_cls_name}` "
            f"and module `{error.exception_module_name}` - fall back to "
            f"`{definitions.GenericRemoteException.__name__}`."
        )
        exception_cls = definitions.GenericRemoteException

    if issubclass(exception_cls, pydantic.ValidationError):
        # Cannot re-raise naively.
        # https://github.com/pydantic/pydantic/issues/6734.
        exception_cls = definitions.GenericRemoteException

    return exception_cls


def _handle_response_error(response_json: dict, remote_name: str):
    try:
        error_json = response_json["error"]
    except KeyError as e:
        logging.error(f"response_json: {response_json}")
        raise ValueError(
            "Could not get `error` field from JSON from chainlet error response"
        ) from e

    try:
        error = definitions.RemoteErrorDetail.model_validate(error_json)
    except pydantic.ValidationError as e:
        if isinstance(error_json, str):
            msg = f"Remote error occurred in `{remote_name}`: '{error_json}'"
            raise definitions.GenericRemoteException(msg) from None
        raise ValueError(
            "Could not parse chainlet error. Error details are expected to be either a "
            "plain string (old truss models) or a serialized "
            f"`{definitions.RemoteErrorDetail.__name__}`, got:\n{repr(error_json)}"
        ) from e

    exception_cls = _resolve_exception_class(error)
    error_format = textwrap.indent(error.format(), "│   ")
    *lines, last_line = error_format.splitlines()
    last_line = f"╰{last_line[1:]}" if last_line.startswith("│") else last_line
    error_format = "\n".join(lines + [last_line])
    msg = (
        f"(showing chained remote errors, root error at the bottom)\n"
        f"├─ Error in dependency Chainlet `{remote_name}`:\n"
        f"{error_format}"
    )
    raise exception_cls(msg)


def response_raise_errors(response: httpx.Response, remote_name: str) -> None:
    """In case of error, raise it.

    If the response error contains `RemoteErrorDetail`, it tries to re-raise
    the same exception that was raised remotely and falls back to
    `GenericRemoteException` if the exception class could not be resolved.

    Exception messages are chained to trace back to the root cause, i.e. the first
    Chainlet that raised an exception. E.g. the message might look like this:

    ```
    RemoteChainletError in "Chain"
    Traceback (most recent call last):
      File "/app/model/Chainlet.py", line 112, in predict
        result = await self._chainlet.run(
      File "/app/model/Chainlet.py", line 79, in run
        value += self._text_to_num.run(part)
      File "/packages/remote_stubs.py", line 21, in run
        json_result = self.predict_sync(json_args)
      File "/packages/truss_chains/stub.py", line 37, in predict_sync
        return utils.handle_response(
    ValueError: (showing remote errors, root message at the bottom)
    --> Preceding Remote Cause:
        RemoteChainletError in "TextToNum"
        Traceback (most recent call last):
          File "/app/model/Chainlet.py", line 113, in predict
            result = self._chainlet.run(data=payload["data"])
          File "/app/model/Chainlet.py", line 54, in run
            generated_text = self._replicator.run(data)
          File "/packages/remote_stubs.py", line 7, in run
            json_result = self.predict_sync(json_args)
          File "/packages/truss_chains/stub.py", line 37, in predict_sync
            return utils.handle_response(
        ValueError: (showing remote errors, root message at the bottom)
        --> Preceding Remote Cause:
            RemoteChainletError in "TextReplicator"
            Traceback (most recent call last):
              File "/app/model/Chainlet.py", line 112, in predict
                result = self._chainlet.run(data=payload["data"])
              File "/app/model/Chainlet.py", line 36, in run
                raise ValueError(f"This input is too long: {len(data)}.")
            ValueError: This input is too long: 100.

    ```
    """
    if response.is_error:
        try:
            response_json = response.json()
        except Exception as e:
            raise ValueError(
                "Could not get JSON from error response. Status: "
                f"`{response.status_code}`."
            ) from e
        _handle_response_error(response_json=response_json, remote_name=remote_name)


async def async_response_raise_errors(
    response: aiohttp.ClientResponse, remote_name: str
) -> None:
    """Async version of `async_response_raise_errors`."""
    if response.status >= 400:
        try:
            response_json = await response.json()
        except Exception as e:
            raise ValueError(
                "Could not get JSON from error response. Status: "
                f"`{response.status}`."
            ) from e
        _handle_response_error(response_json=response_json, remote_name=remote_name)



import pytest
from truss.templates.control.control.helpers.truss_patch.requirement_name_identifier import (
    RequirementMeta,
    identify_requirement_name,
    reqs_by_name,
)


@pytest.mark.parametrize(
    "req, expected_name",
    [
        ("pytorch", "pytorch"),
        (
            "git+https://github.com/huggingface/transformers.git#egg=transformers",
            "git+github.com/huggingface/transformers.git",
        ),
        (
            "git+https://github.com/huggingface/transformers.git",
            "git+github.com/huggingface/transformers.git",
        ),
        (
            "git+https://github.com/huggingface/transformers.git@main#egg=transformers",
            "git+github.com/huggingface/transformers.git",
        ),
        (
            " git+https://github.com/huggingface/transformers.git ",
            "git+github.com/huggingface/transformers.git",
        ),
        ("pytorch==1.0", "pytorch"),
        ("pytorch>=1.0", "pytorch"),
        ("pytorch<=1.0", "pytorch"),
    ],
)
def test_identify_requirement_name(req, expected_name):
    assert expected_name == identify_requirement_name(req)


def test_reqs_by_name():
    reqs = [
        "pytorch",
        " ",
        "jinja==1.0",
    ]
    assert reqs_by_name(reqs) == {"pytorch": "pytorch", "jinja": "jinja==1.0"}


@pytest.mark.parametrize(
    "desc, req, expected_meta",
    [
        (
            "handles simple requirement",
            "pytorch",
            RequirementMeta(
                requirement="pytorch",
                name="pytorch",
                is_url_based_requirement=False,
                egg_tag=None,
            ),
        ),
        (
            "handles python package with version",
            "pytorch==1.0",
            RequirementMeta(
                requirement="pytorch==1.0",
                name="pytorch",
                is_url_based_requirement=False,
                egg_tag=None,
            ),
        ),
        (
            "handles url-based requirement with egg tag",
            "git+https://github.com/huggingface/transformers.git@main#egg=transformers",
            RequirementMeta(
                requirement="git+https://github.com/huggingface/transformers.git@main#egg=transformers",
                name="git+github.com/huggingface/transformers.git",
                is_url_based_requirement=True,
                egg_tag=["transformers"],
            ),
        ),
        (
            "handles url-based requirement without egg tag",
            "git+https://github.com/huggingface/transformers.git",
            RequirementMeta(
                requirement="git+https://github.com/huggingface/transformers.git",
                name="git+github.com/huggingface/transformers.git",
                is_url_based_requirement=True,
                egg_tag=None,
            ),
        ),
    ],
)
def test_requirement_meta_from_req(desc, req: str, expected_meta: RequirementMeta):
    assert expected_meta == RequirementMeta.from_req(req), desc



import copy
from dataclasses import replace
from pathlib import Path
from typing import Optional

from truss.base.validation import validate_secret_name
from truss.local.local_config import LocalConfig


class LocalConfigHandler:
    TRUSS_CONFIG_DIR = Path.home() / ".truss"

    @staticmethod
    def _ensure_config_dir():
        LocalConfigHandler.TRUSS_CONFIG_DIR.mkdir(exist_ok=True, parents=True)

    @staticmethod
    def get_config() -> LocalConfig:
        if LocalConfigHandler._config_path().exists():
            return LocalConfig.from_yaml(LocalConfigHandler._config_path())
        return LocalConfig()

    @staticmethod
    def sync_secrets_mount_dir():
        """Syncs config secrets into a directory form, meant for mounting onto docker containers."""
        LocalConfigHandler._ensure_config_dir()
        local_config = LocalConfigHandler.get_config()
        secrets = local_config.secrets
        secrets_dir = LocalConfigHandler.secrets_dir_path()
        if not secrets_dir.exists():
            secrets_dir.mkdir(parents=True)
        # Remove any files that don't correspond to a secret
        for path in secrets_dir.iterdir():
            if path.is_file() and path.name not in secrets:
                path.unlink()

        for secret_name, secret_value in secrets.items():
            secret_path = secrets_dir / secret_name
            with secret_path.open("w") as secret_file:
                secret_file.write(secret_value)

    @staticmethod
    def set_secret(secret_name: str, secret_value: str):
        LocalConfigHandler._ensure_config_dir()
        validate_secret_name(secret_name)
        local_config = LocalConfigHandler.get_config()
        new_secrets = {
            **local_config.secrets,
            secret_name: secret_value,
        }
        new_local_config = replace(local_config, secrets=new_secrets)
        new_local_config.write_to_yaml_file(LocalConfigHandler._config_path())

    @staticmethod
    def remove_secret(secret_name: str):
        LocalConfigHandler._ensure_config_dir()
        LocalConfigHandler.TRUSS_CONFIG_DIR.mkdir(exist_ok=True, parents=True)
        local_config = LocalConfigHandler.get_config()
        new_secrets = copy.deepcopy(local_config.secrets)
        del new_secrets[secret_name]
        new_local_config = replace(local_config, secrets=new_secrets)
        new_local_config.write_to_yaml_file(LocalConfigHandler._config_path())

    @staticmethod
    def _config_path():
        return LocalConfigHandler.TRUSS_CONFIG_DIR / "config.yaml"

    @staticmethod
    def secrets_dir_path():
        return LocalConfigHandler.TRUSS_CONFIG_DIR / "secrets"

    @staticmethod
    def bptr_data_resolution_dir_path():
        bptr_data_dir = LocalConfigHandler.TRUSS_CONFIG_DIR / "bptr"
        bptr_data_dir.mkdir(exist_ok=True, parents=True)
        return bptr_data_dir

    @staticmethod
    def dynamic_config_path():
        dynamic_config_dir = LocalConfigHandler.TRUSS_CONFIG_DIR / "b10_dynamic_config"
        dynamic_config_dir.mkdir(exist_ok=True, parents=True)
        return dynamic_config_dir

    @staticmethod
    def set_dynamic_config(key: str, value: str):
        key_path = LocalConfigHandler.dynamic_config_path() / key
        with key_path.open("w") as key_file:
            key_file.write(value)

    @staticmethod
    def _signatures_dir_path():
        return LocalConfigHandler.TRUSS_CONFIG_DIR / "signatures"

    @staticmethod
    def shadow_trusses_dir_path():
        return LocalConfigHandler.TRUSS_CONFIG_DIR / "shadow_trusses"

    @staticmethod
    def add_signature(truss_hash: str, signature: str):
        if truss_hash is None:
            raise ValueError("truss_hash is None")

        LocalConfigHandler._ensure_config_dir()
        signature_dir = LocalConfigHandler._signatures_dir_path()
        signature_dir.mkdir(exist_ok=True)
        with (signature_dir / truss_hash).open("w") as signature_file:
            signature_file.write(signature)

    @staticmethod
    def get_signature(truss_hash: str) -> Optional[str]:
        if truss_hash is None:
            raise ValueError("truss_hash is None")

        LocalConfigHandler._ensure_config_dir()
        signature_dir = LocalConfigHandler._signatures_dir_path()
        signature_dir.mkdir(exist_ok=True)
        signature_file_path = signature_dir / truss_hash
        if not signature_file_path.exists() or not signature_file_path.is_file():
            return None

        with signature_file_path.open() as signature_file:
            return signature_file.read()



import subprocess
from pathlib import Path

import pytest


@pytest.mark.integration
def test_build_docker_image(test_data_path):
    root_path = Path(__file__).parent.parent.parent
    root = str(root_path)
    context_builder_image_test_dir = str(test_data_path / "context_builder_image_test")

    subprocess.run(
        [
            "docker",
            "buildx",
            "build",
            ".",
            "-f",
            "context_builder.Dockerfile",
            "--platform=linux/amd64",
            "-t",
            "baseten/truss-context-builder:test",
        ],
        check=True,
        cwd=root,
    )

    subprocess.run(
        [
            "docker",
            "buildx",
            "build",
            context_builder_image_test_dir,
            "--platform=linux/amd64",
            "-t",
            "baseten/truss-context-builder-test",
        ],
        check=True,
        cwd=root,
    )

    # This will throw if building docker build context fails
    subprocess.run(
        [
            "docker",
            "run",
            "baseten/truss-context-builder-test",
        ],
        check=True,
        cwd=root,
    )



import logging
from pathlib import Path
from typing import Dict, List, Optional, Set

import yaml

from truss.base.constants import CONFIG_FILE
from truss.base.truss_config import ExternalData, TrussConfig
from truss.base.truss_spec import TrussSpec
from truss.templates.control.control.helpers.custom_types import (
    Action,
    ConfigPatch,
    DataPatch,
    EnvVarPatch,
    ExternalDataPatch,
    ModelCodePatch,
    PackagePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.templates.control.control.helpers.truss_patch.requirement_name_identifier import (
    RequirementMeta,
)
from truss.templates.control.control.helpers.truss_patch.system_packages import (
    system_packages_set,
)
from truss.truss_handle.patch.custom_types import ChangedPaths, TrussSignature
from truss.truss_handle.patch.hash import file_content_hash_str
from truss.util.path import get_ignored_relative_paths

logger: logging.Logger = logging.getLogger(__name__)
PYCACHE_IGNORE_PATTERNS = [
    "**/__pycache__/**/*",
    "**/__pycache__/**",
]
UNPATCHABLE_CONFIG_KEYS = [
    "live_reload",
    "python_version",
    "resources",
]


def calc_truss_patch(
    truss_dir: Path,
    previous_truss_signature: TrussSignature,
    ignore_patterns: Optional[List[str]] = None,
) -> Optional[List[Patch]]:
    """
    Calculate patch for a truss from a previous state.

    Returns: None if patch cannot be calculated, otherwise a list of patches.
        Note that the none return value is pretty important, patch coverage is
        limited and this usually indicates that the identified change cannot be
        expressed with currently supported patches.

        Only standard and relevant truss paths are scanned for changes, rest
        ignored. E.g. at the root level, only changes to config.yaml are
        checked, any other changes are ignored.
    """

    def _relative_to_root(path: Path) -> str:
        return str(path.relative_to(truss_dir))

    if ignore_patterns is None:
        ignore_patterns = PYCACHE_IGNORE_PATTERNS

    changed_paths = _calc_changed_paths(
        truss_dir,
        previous_truss_signature.content_hashes_by_path,
        ignore_patterns,
    )

    new_config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
    prev_config = TrussConfig.from_dict(yaml.safe_load(previous_truss_signature.config))

    truss_spec = TrussSpec(truss_dir)
    model_module_path = _relative_to_root(truss_spec.model_module_dir)
    data_dir_path = _relative_to_root(truss_spec.data_dir)
    bundled_packages_path = _relative_to_root(truss_spec.bundled_packages_dir)

    def _under_unsupported_patch_dir(path: str) -> bool:
        """
        Checks if the given path is under one of the directories that don't
        support patching. Note that if path `is` one of those directories that's
        ok, because those empty directories can be ignored from patching point
        of view.
        """
        return _strictly_under(path, [data_dir_path])

    patches = []

    for path in changed_paths["removed"]:
        if _strictly_under(path, [model_module_path]):
            logger.info(f"Created patch to remove model code file: {path}")
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=Action.REMOVE,
                        path=_relative_to(path, model_module_path),
                    ),
                )
            )
        elif path == CONFIG_FILE:
            # Don't support removal of config file
            logger.info(f"Patching not supported for removing {path}")
            return None
        elif _strictly_under(path, [bundled_packages_path]):
            logger.info(f"Created patch to remove package file: {path}")
            patches.append(
                Patch(
                    type=PatchType.PACKAGE,
                    body=PackagePatch(
                        action=Action.REMOVE,
                        path=_relative_to(path, bundled_packages_path),
                    ),
                )
            )
        elif _under_unsupported_patch_dir(path):
            logger.warning(f"Patching not supported for removing {path}")
            return None

    has_calculated_config = False
    for path in changed_paths["added"] + changed_paths["updated"]:
        action = Action.ADD if path in changed_paths["added"] else Action.UPDATE
        if _strictly_under(path, [model_module_path]):
            full_path = truss_dir / path

            # TODO(pankaj) Add support for empty directories, skip them for now.
            if not full_path.is_file():
                continue
            logger.info(
                f"Created patch to {action.value.lower()} model code file: {path}"
            )
            patches.append(
                Patch(
                    type=PatchType.MODEL_CODE,
                    body=ModelCodePatch(
                        action=action,
                        path=_relative_to(path, model_module_path),
                        content=full_path.read_text(),
                    ),
                )
            )
        elif (
            path == CONFIG_FILE or _changed_path_is_requirements_file(path, new_config)
        ) and not has_calculated_config:
            # we could enter this code block from the requirements file's path
            # or from the config file's path. In any case, we only want to calculate these
            # patches once.
            has_calculated_config = True
            config_patches = _calc_config_patches(
                truss_dir, previous_truss_signature, prev_config, new_config
            )
            if config_patches:
                logger.info(f"Created patch to {action.value.lower()} config")
            patches.extend(config_patches)
        elif _strictly_under(path, [bundled_packages_path]):
            full_path = truss_dir / path
            if not full_path.is_file():
                continue
            logger.info(f"Created patch to {action.value.lower()} package file: {path}")
            patches.append(
                Patch(
                    type=PatchType.PACKAGE,
                    body=PackagePatch(
                        action=action,
                        path=_relative_to(path, bundled_packages_path),
                        content=full_path.read_text(),
                    ),
                )
            )
        elif _under_unsupported_patch_dir(path):
            logger.info(f"Patching not supported for updating {path}")
            return None
    return patches


def _changed_path_is_requirements_file(changed_path: str, new_config: TrussConfig):
    """
    _changed_path_is_requirements_file determines if `changed_path` is the same path
    as the requirements file on a new configuration.
    """
    return new_config.requirements_file and Path(changed_path) == Path(
        new_config.requirements_file
    )


def _calc_changed_paths(
    root: Path,
    previous_root_path_content_hashes: Dict[str, str],
    ignore_patterns: Optional[List[str]],
) -> ChangedPaths:
    """
    TODO(pankaj) add support for directory creation in patch
    """
    root_relative_new_paths = set(
        (str(path.relative_to(root)) for path in root.glob("**/*"))
    )
    unignored_new_paths = _calc_unignored_paths(
        root_relative_new_paths, ignore_patterns
    )
    previous_root_relative_paths = set(previous_root_path_content_hashes.keys())
    unignored_prev_paths = _calc_unignored_paths(
        previous_root_relative_paths, ignore_patterns
    )

    added_paths = unignored_new_paths - unignored_prev_paths
    removed_paths = unignored_prev_paths - unignored_new_paths

    updated_paths = set()
    common_paths = unignored_new_paths.intersection(unignored_prev_paths)
    for path in common_paths:
        full_path: Path = root / path
        if full_path.is_file():
            content_hash = file_content_hash_str(full_path)
            previous_content_hash = previous_root_path_content_hashes[path]
            if content_hash != previous_content_hash:
                updated_paths.add(path)

    return {
        "added": list(added_paths),
        "updated": list(updated_paths),
        "removed": list(removed_paths),
    }


def _calc_unignored_paths(
    root_relative_paths: Set[str],
    ignore_patterns: Optional[List[str]] = None,
) -> Set[str]:
    ignored_paths = set(
        get_ignored_relative_paths(root_relative_paths, ignore_patterns)
    )
    return root_relative_paths - ignored_paths  # type: ignore


def _calc_config_patches(
    truss_dir: Path,
    prev_signature: TrussSignature,
    prev_config: TrussConfig,
    new_config: TrussConfig,
) -> List[Patch]:
    """Calculate patch based on changes to config.

    Returns None if patch cannot be calculated. Empty list means no relevant
    differences found.
    """
    try:
        config_patches = _calc_general_config_patches(prev_config, new_config)
        python_requirements_patches = _calc_requirements_patches(
            truss_dir, prev_signature, prev_config, new_config
        )
        system_package_patches = _calc_system_packages_patches(prev_config, new_config)
        return [*config_patches, *python_requirements_patches, *system_package_patches]
    except Exception as e:
        logger.error(f"Failed to calculate config patch with exception: {e}")
        raise


def _calc_general_config_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to config.yaml
    If a change has been made to the config, at least one ConfigPatch is created
    Additional patches are created for each of the patches that need specific application,
    namely patches to the env variables, external package dirs, and/or external data.

    Empty list means no relevant differences found.
    """

    patches: List[Patch] = []
    for key in UNPATCHABLE_CONFIG_KEYS:
        prev_items = getattr(prev_config, key)
        new_items = getattr(new_config, key)
        if new_items != prev_items:
            logger.warning(f"Patching is not supported for: {key}")

    if prev_config.to_dict() != new_config.to_dict():
        patches.append(_mk_config_patch(Action.UPDATE, new_config.to_dict()))

    env_var_patches = _calc_env_var_patches(prev_config, new_config)
    external_data_patches = _calc_external_data_patches(prev_config, new_config)

    return [*patches, *env_var_patches, *external_data_patches]


def _calc_env_var_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to environment variables.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_items = prev_config.environment_variables
    new_items = new_config.environment_variables
    prev_item_names = set(prev_items.keys())
    new_item_names = set(new_items.keys())
    removed_items = prev_item_names.difference(new_item_names)
    for removed_item in removed_items:
        patches.append(
            _mk_env_var_patch(
                Action.REMOVE,
                {removed_item: prev_items[removed_item]},
            )
        )
    added_items = new_item_names.difference(prev_item_names)
    for added_item in added_items:
        patches.append(
            _mk_env_var_patch(Action.ADD, {added_item: new_items[added_item]})
        )
    for item in new_item_names.intersection(prev_item_names):
        if new_items[item] != prev_items[item]:
            patches.append(_mk_env_var_patch(Action.UPDATE, {item: new_items[item]}))
    return patches


def _calc_external_data_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to external data.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_items = (prev_config.external_data or ExternalData([])).to_list()
    new_items = (new_config.external_data or ExternalData([])).to_list()

    removed_items = [x for x in prev_items if x not in new_items]
    for removed_item in removed_items:
        patches.append(
            _mk_external_data_patch(
                Action.REMOVE,
                removed_item,
            )
        )
    added_items = [x for x in new_items if x not in prev_items]
    for added_item in added_items:
        patches.append(_mk_external_data_patch(Action.ADD, added_item))
    return patches


def _calc_requirements_patches(
    truss_dir: Path,
    prev_signature: TrussSignature,
    prev_config: TrussConfig,
    new_config: TrussConfig,
) -> List[Patch]:
    """
    requirements patches that are accounted for
    - contents in the requirements file change
    - requirements move from config file to requirements file
    - requirements move from requirements file to config file
    - requirement file changes (i.e. requirement.txt --> requirements.txt)
    - requirements in the config file change

    assumes that only one of requirements or requirements_file is present for a given config
    """
    prev_requirements = (
        prev_config.requirements
        if not prev_config.requirements_file
        else prev_signature.requirements_file_requirements
    )
    new_requirements = (
        new_config.requirements
        if not new_config.requirements_file
        else new_config.load_requirements_from_file(truss_dir)
    )
    return _calc_python_requirements_patches(prev_requirements, new_requirements)


def _calc_python_requirements_patches(
    prev_raw_reqs: List[str], new_raw_reqs: List[str]
) -> List[Patch]:
    """Calculate patch based on changes to python requirements.

    Empty list means no relevant differences found.
    """
    patches = []

    def create_requirement_map(raw_reqs: List[str]) -> Dict[str, RequirementMeta]:
        req_map = {}
        for raw_req in raw_reqs:
            meta = RequirementMeta.from_req(raw_req)
            req_map[meta.name] = meta
        return req_map

    prev_reqs_map = create_requirement_map(prev_raw_reqs)
    new_reqs_map = create_requirement_map(new_raw_reqs)
    prev_req_names = set(prev_reqs_map.keys())
    new_req_names = set(new_reqs_map.keys())

    removed_req_names = prev_req_names.difference(new_req_names)
    for removed_req_name in removed_req_names:
        removed_req_meta = prev_reqs_map[removed_req_name]
        if removed_req_meta.is_url_based_requirement:
            if not removed_req_meta.egg_tag:
                logger.warning(
                    f"Url-based requirement `{removed_req_meta.requirement}` is missing egg tag. Ignoring removal. Use `truss push` if you want to remove this requirement."
                )
                continue
            # make sure that the egg tag is included when removing the requirement
            patches.append(
                _mk_python_requirement_patch(
                    Action.REMOVE, removed_req_meta.requirement
                )
            )
        else:
            patches.append(
                _mk_python_requirement_patch(Action.REMOVE, removed_req_meta.name)
            )

    # warn for new reqs
    added_req_names = new_req_names.difference(prev_req_names)
    for added_req_name in added_req_names:
        added_req_meta = new_reqs_map[added_req_name]
        patches.append(
            _mk_python_requirement_patch(Action.ADD, added_req_meta.requirement)
        )
        if added_req_meta.is_url_based_requirement and not added_req_meta.egg_tag:
            logger.warning(
                f"Url-based requirement `{added_req_meta.requirement}` is missing egg tag. Removal will be ignored by `truss watch`"
            )
    for req_name in new_req_names.intersection(prev_req_names):
        if prev_reqs_map[req_name].requirement != new_reqs_map[req_name].requirement:
            patches.append(
                _mk_python_requirement_patch(
                    Action.UPDATE, new_reqs_map[req_name].requirement
                )
            )

    return patches


def _calc_system_packages_patches(
    prev_config: TrussConfig, new_config: TrussConfig
) -> List[Patch]:
    """Calculate patch based on changes to system packates.

    Empty list means no relevant differences found.
    """
    patches = []
    prev_pkgs = system_packages_set(prev_config.system_packages)
    new_pkgs = system_packages_set(new_config.system_packages)
    removed_pkgs = prev_pkgs.difference(new_pkgs)
    for removed_pkg in removed_pkgs:
        patches.append(_mk_system_package_patch(Action.REMOVE, removed_pkg))

    added_pkgs = new_pkgs.difference(prev_pkgs)
    for added_pkg in added_pkgs:
        patches.append(_mk_system_package_patch(Action.ADD, added_pkg))

    return patches


def _mk_config_patch(action: Action, config: dict) -> Patch:
    return Patch(
        type=PatchType.CONFIG,
        body=ConfigPatch(
            action=action,
            config=config,
        ),
    )


# Support for patching data changes yet to be implemented
def _mk_data_patch(action: Action, item: str, path: str) -> Patch:
    return Patch(
        type=PatchType.DATA,
        body=DataPatch(action=action, content=item, path=path),
    )


def _mk_env_var_patch(action: Action, item: dict) -> Patch:
    return Patch(
        type=PatchType.ENVIRONMENT_VARIABLE,
        body=EnvVarPatch(
            action=action,
            item=item,
        ),
    )


def _mk_external_data_patch(action: Action, item: Dict[str, str]) -> Patch:
    return Patch(
        type=PatchType.EXTERNAL_DATA,
        body=ExternalDataPatch(
            action=action,
            item=item,
        ),
    )


def _mk_python_requirement_patch(action: Action, requirement: str) -> Patch:
    return Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=action,
            requirement=requirement,
        ),
    )


def _mk_system_package_patch(action: Action, package: str) -> Patch:
    return Patch(
        type=PatchType.SYSTEM_PACKAGE,
        body=SystemPackagePatch(
            action=action,
            package=package,
        ),
    )


def _relative_to(path: str, relative_to_path: str):
    return str(Path(path).relative_to(relative_to_path))


def _strictly_under(path: str, parent_paths: List[str]) -> bool:
    """
    Checks if given path is under one of the given paths, but not the same as
    them. Assumes that parent paths themselves are not under each other.
    """
    for dir_path in parent_paths:
        if path.startswith(dir_path) and not path == dir_path:
            return True
    return False



import abc
import asyncio
import contextlib
import contextvars
import json
import logging
import threading
import time
from typing import (
    Any,
    AsyncIterator,
    ClassVar,
    Dict,
    Iterator,
    Mapping,
    Optional,
    Type,
    TypeVar,
    Union,
    final,
    overload,
)

import aiohttp
import httpx
import pydantic
import starlette.requests
import tenacity
from truss.templates.shared import serialization

from truss_chains import definitions
from truss_chains.remote_chainlet import utils

DEFAULT_MAX_CONNECTIONS = 1000
DEFAULT_MAX_KEEPALIVE_CONNECTIONS = 400


_RetryPolicyT = TypeVar("_RetryPolicyT", tenacity.AsyncRetrying, tenacity.Retrying)
_InputT = TypeVar("_InputT", pydantic.BaseModel, Any)  # Any signifies "JSON".
_OutputT = TypeVar("_OutputT", bound=pydantic.BaseModel)


_trace_parent_context: contextvars.ContextVar[str] = contextvars.ContextVar(
    "trace_parent"
)


@contextlib.contextmanager
def trace_parent(request: starlette.requests.Request) -> Iterator[None]:
    token = _trace_parent_context.set(
        request.headers.get(definitions.OTEL_TRACE_PARENT_HEADER_KEY, "")
    )
    try:
        yield
    finally:
        _trace_parent_context.reset(token)


class BasetenSession:
    """Provides configured HTTP clients, retries rate limit warning etc."""

    _client_cycle_time_sec: ClassVar[int] = 3600 * 1  # 1 hour.
    _client_limits: ClassVar[httpx.Limits] = httpx.Limits(
        max_connections=DEFAULT_MAX_CONNECTIONS,
        max_keepalive_connections=DEFAULT_MAX_KEEPALIVE_CONNECTIONS,
    )
    _auth_header: Mapping[str, str]
    _service_descriptor: definitions.DeployedServiceDescriptor
    _cached_sync_client: Optional[tuple[httpx.Client, int]]
    _cached_async_client: Optional[tuple[aiohttp.ClientSession, int]]

    def __init__(
        self,
        service_descriptor: definitions.DeployedServiceDescriptor,
        api_key: str,
    ) -> None:
        logging.info(
            f"Creating BasetenSession (HTTP) for `{service_descriptor.name}`.\n"
            f"\tTarget: `{service_descriptor.predict_url}`\n"
            f"\t`{service_descriptor.options}`."
        )
        self._auth_header = {"Authorization": f"Api-Key {api_key}"}
        self._service_descriptor = service_descriptor
        self._cached_sync_client = None
        self._cached_async_client = None
        self._sync_lock = threading.Lock()
        self._async_lock = asyncio.Lock()
        self._sync_num_requests = utils.ThreadSafeCounter()
        self._async_num_requests = utils.AsyncSafeCounter()

    @property
    def name(self) -> str:
        return self._service_descriptor.name

    def _maybe_warn_for_overload(self, num_requests: int) -> None:
        if self._client_limits.max_connections is None:
            return
        if num_requests > self._client_limits.max_connections * 0.8:
            logging.warning(
                f"High number of concurrently outgoing HTTP connections: "
                f"`{num_requests}`. Close to or above connection limit of "
                f"`{self._client_limits.max_connections}`. To avoid overload and "
                f"timeouts, use more replicas/autoscaling for this chainlet."
            )

    def _client_cycle_needed(self, cached_client: Optional[tuple[Any, int]]) -> bool:
        return (
            not cached_client
            or (int(time.time()) - cached_client[1]) > self._client_cycle_time_sec
        )

    def _log_retry(self, retry_state: tenacity.RetryCallState) -> None:
        logging.info(f"Retrying `{self.name}`, attempt {retry_state.attempt_number}")

    def _make_retry_policy(self, retrying: Type[_RetryPolicyT]) -> _RetryPolicyT:
        return retrying(
            stop=tenacity.stop_after_attempt(self._service_descriptor.options.retries),
            retry=tenacity.retry_if_exception_type(Exception),
            reraise=True,
            before_sleep=self._log_retry,
        )

    @contextlib.contextmanager
    def _client_sync(self) -> Iterator[httpx.Client]:
        # Check `_client_cycle_needed` before and after locking to avoid
        # needing a lock each time the client is accessed.
        if self._client_cycle_needed(self._cached_sync_client):
            with self._sync_lock:
                if self._client_cycle_needed(self._cached_sync_client):
                    self._cached_sync_client = (
                        httpx.Client(
                            headers=self._auth_header,
                            timeout=self._service_descriptor.options.timeout_sec,
                            limits=self._client_limits,
                        ),
                        int(time.time()),
                    )
        assert self._cached_sync_client is not None
        client = self._cached_sync_client[0]

        with self._sync_num_requests as num_requests:
            self._maybe_warn_for_overload(num_requests)
            yield client

    @contextlib.asynccontextmanager
    async def _client_async(self) -> AsyncIterator[aiohttp.ClientSession]:
        # Check `_client_cycle_needed` before and after locking to avoid
        # needing a lock each time the client is accessed.
        if self._client_cycle_needed(self._cached_async_client):
            async with self._async_lock:
                if self._client_cycle_needed(self._cached_async_client):
                    connector = aiohttp.TCPConnector(
                        limit=DEFAULT_MAX_CONNECTIONS,
                    )
                    self._cached_async_client = (
                        aiohttp.ClientSession(
                            headers=self._auth_header,
                            connector=connector,
                            timeout=aiohttp.ClientTimeout(
                                total=self._service_descriptor.options.timeout_sec
                            ),
                        ),
                        int(time.time()),
                    )
        assert self._cached_async_client is not None
        client = self._cached_async_client[0]

        async with self._async_num_requests as num_requests:
            self._maybe_warn_for_overload(num_requests)
            yield client


class StubBase(BasetenSession, abc.ABC):
    """Base class for stubs that invoke remote chainlets.

    Extends ``BasetenSession`` with methods for data serialization, de-serialization
    and invoking other endpoints.

    It is used internally for RPCs to dependency chainlets, but it can also be used
    in user-code for wrapping a deployed truss model into the chains framework, e.g.
    like that::

        import pydantic
        import truss_chains as chains

        class WhisperOutput(pydantic.BaseModel):
            ...


        class DeployedWhisper(chains.StubBase):

            async def run_remote(self, audio_b64: str) -> WhisperOutput:
                resp = await self.predict_async(
                    json_payload={"audio": audio_b64})
                return WhisperOutput(text=resp["text"], language=resp["language"])


        class MyChainlet(chains.ChainletBase):

            def __init__(self, ..., context=chains.depends_context()):
                ...
                self._whisper = DeployedWhisper.from_url(
                    WHISPER_URL,
                    context,
                    options=chains.RPCOptions(retries=3),
                )

    """

    @final
    def __init__(
        self,
        service_descriptor: definitions.DeployedServiceDescriptor,
        api_key: str,
    ) -> None:
        """
        Args:
            service_descriptor: Contains the URL and other configuration.
            api_key: A baseten API key to authorize requests.
        """
        super().__init__(service_descriptor, api_key)

    @classmethod
    def from_url(
        cls,
        predict_url: str,
        context: definitions.DeploymentContext,
        options: Optional[definitions.RPCOptions] = None,
    ):
        """Factory method, convenient to be used in chainlet's ``__init__``-method.

        Args:
            predict_url: URL to predict endpoint of another chain / truss model.
            context: Deployment context object, obtained in the chainlet's ``__init__``.
            options: RPC options, e.g. retries.
        """
        options = options or definitions.RPCOptions()
        return cls(
            service_descriptor=definitions.DeployedServiceDescriptor(
                name=cls.__name__,
                display_name=cls.__name__,
                predict_url=predict_url,
                options=options,
            ),
            api_key=context.get_baseten_api_key(),
        )

    def _make_request_params(
        self, inputs: _InputT, for_httpx: bool = False
    ) -> Mapping[str, Any]:
        kwargs: Dict[str, Any] = {}
        headers = {
            definitions.OTEL_TRACE_PARENT_HEADER_KEY: _trace_parent_context.get()
        }
        if isinstance(inputs, pydantic.BaseModel):
            if self._service_descriptor.options.use_binary:
                data_dict = inputs.model_dump(mode="python")
                kwargs["data"] = serialization.truss_msgpack_serialize(data_dict)
                headers["Content-Type"] = "application/octet-stream"
            else:
                data_key = "content" if for_httpx else "data"
                kwargs[data_key] = inputs.model_dump_json()
                headers["Content-Type"] = "application/json"
        else:  # inputs is JSON dict.
            if self._service_descriptor.options.use_binary:
                kwargs["data"] = serialization.truss_msgpack_serialize(inputs)
                headers["Content-Type"] = "application/octet-stream"
            else:
                kwargs["json"] = inputs
                headers["Content-Type"] = "application/json"

        kwargs["headers"] = headers
        return kwargs

    def _response_to_pydantic(
        self, response: bytes, output_model: Type[_OutputT]
    ) -> _OutputT:
        if self._service_descriptor.options.use_binary:
            data_dict = serialization.truss_msgpack_deserialize(response)
            return output_model.model_validate(data_dict)
        return output_model.model_validate_json(response)

    def _response_to_json(self, response: bytes) -> Any:
        if self._service_descriptor.options.use_binary:
            return serialization.truss_msgpack_deserialize(response)
        return json.loads(response)

    @overload
    def predict_sync(
        self, inputs: _InputT, output_model: Type[_OutputT]
    ) -> _OutputT: ...

    @overload  # Returns JSON
    def predict_sync(self, inputs: _InputT, output_model: None = None) -> Any: ...

    def predict_sync(
        self, inputs: _InputT, output_model: Optional[Type[_OutputT]] = None
    ) -> Union[_OutputT, Any]:
        retry = self._make_retry_policy(tenacity.Retrying)
        params = self._make_request_params(inputs, for_httpx=True)

        def _rpc() -> bytes:
            client: httpx.Client
            with self._client_sync() as client:
                response = client.post(self._service_descriptor.predict_url, **params)
            utils.response_raise_errors(response, self.name)
            return response.content

        response_bytes = retry(_rpc)
        if output_model:
            return self._response_to_pydantic(response_bytes, output_model)
        return self._response_to_json(response_bytes)

    @overload
    async def predict_async(
        self, inputs: _InputT, output_model: Type[_OutputT]
    ) -> _OutputT: ...

    @overload  # Returns JSON.
    async def predict_async(
        self, inputs: _InputT, output_model: None = None
    ) -> Any: ...

    async def predict_async(
        self, inputs: _InputT, output_model: Optional[Type[_OutputT]] = None
    ) -> Union[_OutputT, Any]:
        retry = self._make_retry_policy(tenacity.AsyncRetrying)
        params = self._make_request_params(inputs)

        async def _rpc() -> bytes:
            client: aiohttp.ClientSession
            async with self._client_async() as client:
                async with client.post(
                    self._service_descriptor.predict_url, **params
                ) as response:
                    await utils.async_response_raise_errors(response, self.name)
                    return await response.read()

        response_bytes: bytes = await retry(_rpc)
        if output_model:
            return self._response_to_pydantic(response_bytes, output_model)
        return self._response_to_json(response_bytes)

    async def predict_async_stream(self, inputs: _InputT) -> AsyncIterator[bytes]:
        retry = self._make_retry_policy(tenacity.AsyncRetrying)
        params = self._make_request_params(inputs)

        async def _rpc() -> AsyncIterator[bytes]:
            client: aiohttp.ClientSession
            async with self._client_async() as client:
                response = await client.post(
                    self._service_descriptor.predict_url, **params
                )
                await utils.async_response_raise_errors(response, self.name)
                return response.content.iter_any()

        return await retry(_rpc)


StubT = TypeVar("StubT", bound=StubBase)


def factory(stub_cls: Type[StubT], context: definitions.DeploymentContext) -> StubT:
    # Assumes the stub_cls-name and the name of the service in ``context` match.
    return stub_cls(
        service_descriptor=context.get_service_descriptor(stub_cls.__name__),
        api_key=context.get_baseten_api_key(),
    )



"""
Chains currently assumes that everything from the directory in which the entrypoint
is defined (i.e. sibling files and nested dirs) could be imported/used. e.g.:

workspace/
  entrypoint.py
  helper.py
  some_package/
    utils.py
    sub_package/
      ...

These sources are copied into truss's `/packages` and can be imported on the remote.
Using code *outside* of the workspace is not supported:

shared_lib/
  common.py
workspace/
  entrypoint.py
  ...

`shared_lib` can only be imported on the remote if its installed as a pip
requirement (site-package), it will not be copied from the local host.
"""

import logging
import os
import pathlib
import re
import shlex
import shutil
import subprocess
import sys
import textwrap
from typing import Any, Iterable, Mapping, Optional

import libcst
import truss
from truss.base import truss_config
from truss.contexts.image_builder import serving_image_builder
from truss.util import path as truss_path

from truss_chains import definitions, framework, utils

_INDENT = " " * 4
_REQUIREMENTS_FILENAME = "pip_requirements.txt"
_MODEL_FILENAME = "model.py"
_MODEL_CLS_NAME = "TrussChainletModel"
_TRUSS_GIT = "git+https://github.com/basetenlabs/truss.git"
_TRUSS_PIP_PATTERN = re.compile(
    r"""
    ^truss
    (?:
        \s*(==|>=|<=|!=|>|<)\s*   # Version comparison operators
        \d+(\.\d+)*               # Version numbers (e.g., 1, 1.0, 1.0.0)
        (?:                       # Optional pre-release or build metadata
            (?:a|b|rc|dev)\d*
            (?:\.post\d+)?
            (?:\+[\w\.]+)?
        )?
    )?$
""",
    re.VERBOSE,
)

_MODEL_SKELETON_FILE = (
    pathlib.Path(__file__).parent.parent.resolve()
    / "remote_chainlet"
    / "model_skeleton.py"
)


def _indent(text: str, num: int = 1) -> str:
    return textwrap.indent(text, _INDENT * num)


def _run_simple_subprocess(cmd: str) -> None:
    process = subprocess.Popen(
        shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    _, stderr = process.communicate()
    if process.returncode != 0:
        raise ChildProcessError(f"Error: {stderr.decode()}")


def _format_python_file(file_path: pathlib.Path) -> None:
    # Resolve importing sorting and unused import issues.
    _run_simple_subprocess(f"ruff check {file_path} --fix --select F401,I")
    _run_simple_subprocess(f"ruff format {file_path}")


class _Source(definitions.SafeModelNonSerializable):
    src: str
    imports: set[str] = set()


def _update_src(new_source: _Source, src_parts: list[str], imports: set[str]) -> None:
    src_parts.append(new_source.src)
    imports.update(new_source.imports)


def _gen_pydantic_import_and_ref(raw_type: Any) -> _Source:
    """Returns e.g. ("from sub_package import module", "module.OutputType")."""
    if raw_type.__module__ == "__main__":
        # TODO: assuming that main is copied into package dir and can be imported.
        module_obj = sys.modules[raw_type.__module__]
        if not module_obj.__file__:
            raise definitions.ChainsUsageError(
                f"File-based python code required. `{raw_type}` does not have a file."
            )

        file = os.path.basename(module_obj.__file__)
        assert file.endswith(".py")
        module_name = file.replace(".py", "")
        import_src = f"import {module_name}"
        ref_src = f"{module_name}.{raw_type.__name__}"
    else:
        parts = raw_type.__module__.split(".")
        ref_src = f"{parts[-1]}.{raw_type.__name__}"
        if len(parts) > 1:
            import_src = f"from {'.'.join(parts[:-1])} import {parts[-1]}"
        else:
            import_src = f"import {parts[0]}"

    return _Source(src=ref_src, imports={import_src})


def _gen_type_import_and_ref(type_descr: definitions.TypeDescriptor) -> _Source:
    """Returns e.g. ("from sub_package import module", "module.OutputType")."""
    if type_descr.is_pydantic:
        return _gen_pydantic_import_and_ref(type_descr.raw)

    elif isinstance(type_descr.raw, type):
        if not type_descr.raw.__module__ == "builtins":
            raise TypeError(
                f"{type_descr.raw} is not a builtin - cannot be rendered as source."
            )
        return _Source(src=type_descr.raw.__name__)
    else:
        return _Source(src=str(type_descr.raw))


def _gen_streaming_type_import_and_ref(
    stream_type: definitions.StreamingTypeDescriptor,
) -> _Source:
    """Unlike other `_gen`-helpers, this does not define a type, it creates a symbol."""
    mod = stream_type.origin_type.__module__
    arg = stream_type.arg_type.__name__
    type_src = f"{mod}.{stream_type.origin_type.__name__}[{arg}]"
    return _Source(src=type_src, imports={f"import {mod}"})


def _gen_chainlet_import_and_ref(
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
) -> _Source:
    """Returns e.g. ("from sub_package import module", "module.OutputType")."""
    return _gen_pydantic_import_and_ref(chainlet_descriptor.chainlet_cls)


# I/O used by Stubs and Truss models ###################################################


def _get_input_model_name(chainlet_name: str) -> str:
    return f"{chainlet_name}Input"


def _get_output_model_name(chainlet_name: str) -> str:
    return f"{chainlet_name}Output"


def _gen_truss_input_pydantic(
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
) -> _Source:
    imports = {"import pydantic", "from typing import Optional"}
    fields = []
    for arg in chainlet_descriptor.endpoint.input_args:
        type_ref = _gen_type_import_and_ref(arg.type)
        imports.update(type_ref.imports)
        if arg.is_optional:
            fields.append(f"{arg.name}: Optional[{type_ref.src}] = None")
        else:
            fields.append(f"{arg.name}: {type_ref.src}")

    if fields:
        field_block = _indent("\n".join(fields))
    else:
        field_block = _indent("pass")

    model_name = _get_input_model_name(chainlet_descriptor.name)
    src = f"class {model_name}(pydantic.BaseModel):\n{field_block}"
    return _Source(src=src, imports=imports)


def _gen_truss_output_pydantic(
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
) -> _Source:
    imports = {"import pydantic"}
    fields: list[str] = []
    for i, output_type in enumerate(chainlet_descriptor.endpoint.output_types):
        _update_src(_gen_type_import_and_ref(output_type), fields, imports)

    model_name = _get_output_model_name(chainlet_descriptor.name)
    if len(fields) > 1:
        root_type = f"tuple[{','.join(fields)}]"
    else:
        root_type = fields[0]
    src = f"{model_name} = pydantic.RootModel[{root_type}]"
    return _Source(src=src, imports=imports)


# Stub Gen #############################################################################


def _stub_endpoint_signature_src(
    endpoint: definitions.EndpointAPIDescriptor,
) -> _Source:
    """
    E.g.:
    ```
    async def run_remote(
        self, inputs: shared_chainlet.SplitTextInput, extra_arg: int
    ) -> tuple[shared_chainlet.SplitTextOutput, int]:
    ```
    """
    imports = set()
    args = ["self"]
    for arg in endpoint.input_args:
        arg_ref = _gen_type_import_and_ref(arg.type)
        imports.update(arg_ref.imports)
        args.append(f"{arg.name}: {arg_ref.src}")

    if endpoint.is_streaming:
        streaming_src = _gen_streaming_type_import_and_ref(endpoint.streaming_type)
        imports.update(streaming_src.imports)
        output = streaming_src.src
    else:
        outputs: list[str] = []
        for output_type in endpoint.output_types:
            _update_src(_gen_type_import_and_ref(output_type), outputs, imports)

        if len(outputs) == 1:
            output = outputs[0]
        else:
            output = f"tuple[{', '.join(outputs)}]"

    def_str = "async def" if endpoint.is_async else "def"
    return _Source(
        src=f"{def_str} {endpoint.name}({','.join(args)}) -> {output}:",
        imports=imports,
    )


def _stub_endpoint_body_src(
    endpoint: definitions.EndpointAPIDescriptor, chainlet_name: str
) -> _Source:
    """Generates source code for calling the stub and wrapping the I/O types.

    E.g.:
    ```
    return await self.predict_async(
        SplitTextInput(inputs=inputs, extra_arg=extra_arg), SplitTextOutput).root
    ```
    """
    imports: set[str] = set()
    args = [f"{arg.name}={arg.name}" for arg in endpoint.input_args]
    if args:
        inputs = f"{_get_input_model_name(chainlet_name)}({', '.join(args)})"
    else:
        inputs = "{}"

    parts = []
    # Invoke remote.
    if not endpoint.is_streaming:
        output_model_name = _get_output_model_name(chainlet_name)
        if endpoint.is_async:
            parts = [
                f"return (await self.predict_async({inputs}, {output_model_name})).root"
            ]
        else:
            parts = [f"return self.predict_sync({inputs}, {output_model_name}).root"]

    else:
        if endpoint.is_async:
            parts.append(
                f"async for data in await self.predict_async_stream({inputs}):",
            )
            if endpoint.streaming_type.is_string:
                parts.append(_indent("yield data.decode()"))
            else:
                parts.append(_indent("yield data"))
        else:
            raise NotImplementedError(
                "`Streaming endpoints (containing `yield` statements) are only "
                "supported for async endpoints."
            )

    return _Source(src="\n".join(parts), imports=imports)


def _gen_stub_src(chainlet: definitions.ChainletAPIDescriptor) -> _Source:
    """Generates stub class source, e.g:

    ```
    <IMPORTS>

    class SplitTextInput(pydantic.BaseModel):
        inputs: shared_chainlet.SplitTextInput
        extra_arg: int

    class SplitTextOutput(pydantic.BaseModel):
        output: tuple[shared_chainlet.SplitTextOutput, int]

    class SplitText(stub.StubBase):
        async def run_remote(
            self, inputs: shared_chainlet.SplitTextInput, extra_arg: int
        ) -> tuple[shared_chainlet.SplitTextOutput, int]:
            return await self.predict_async(
                SplitTextInput(inputs=inputs, extra_arg=extra_arg), SplitTextOutput).root
    ```
    """
    imports = {"from truss_chains.remote_chainlet import stub"}
    src_parts: list[str] = []
    input_src = _gen_truss_input_pydantic(chainlet)
    _update_src(input_src, src_parts, imports)
    if not chainlet.endpoint.is_streaming:
        output_src = _gen_truss_output_pydantic(chainlet)
        _update_src(output_src, src_parts, imports)
    signature = _stub_endpoint_signature_src(chainlet.endpoint)
    imports.update(signature.imports)
    body = _stub_endpoint_body_src(chainlet.endpoint, chainlet.name)
    imports.update(body.imports)

    src_parts.extend(
        [
            f"class {chainlet.name}(stub.StubBase):",
            _indent(signature.src),
            _indent(body.src, 2),
            "\n",
        ]
    )
    return _Source(src="\n".join(src_parts), imports=imports)


def _gen_stub_src_for_deps(
    dependencies: Iterable[definitions.ChainletAPIDescriptor],
) -> Optional[_Source]:
    """Generates a source code and imports for stub classes."""
    imports: set[str] = set()
    src_parts: list[str] = []
    for dep in dependencies:
        _update_src(_gen_stub_src(dep), src_parts, imports)

    if not (imports or src_parts):
        return None
    return _Source(src="\n".join(src_parts), imports=imports)


# Truss Chainlet Gen ###################################################################


def _make_chainlet_dir(
    chain_name: str,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    root: pathlib.Path,
) -> pathlib.Path:
    dir_name = f"chainlet_{chainlet_descriptor.name}"
    chainlet_dir = root / definitions.GENERATED_CODE_DIR / chain_name / dir_name
    if chainlet_dir.exists():
        shutil.rmtree(chainlet_dir)
    chainlet_dir.mkdir(exist_ok=False, parents=True)
    return chainlet_dir


class _SpecifyChainletTypeAnnotation(libcst.CSTTransformer):
    """Inserts the concrete chainlet class into `_chainlet: definitions.ABCChainlet`."""

    def __init__(self, new_annotation: str) -> None:
        super().__init__()
        self._new_annotation = libcst.parse_expression(new_annotation)

    def leave_SimpleStatementLine(
        self,
        original_node: libcst.SimpleStatementLine,
        updated_node: libcst.SimpleStatementLine,
    ) -> libcst.SimpleStatementLine:
        new_body: list[Any] = []
        for statement in updated_node.body:
            if (
                isinstance(statement, libcst.AnnAssign)
                and isinstance(statement.target, libcst.Name)
                and statement.target.value == "_chainlet"
            ):
                new_annotation = libcst.Annotation(annotation=self._new_annotation)
                new_statement = statement.with_changes(annotation=new_annotation)
                new_body.append(new_statement)
            else:
                new_body.append(statement)

        return updated_node.with_changes(body=tuple(new_body))


def _gen_load_src(chainlet_descriptor: definitions.ChainletAPIDescriptor) -> _Source:
    """Generates AST for the `load` method of the truss model."""
    imports = {"from truss_chains.remote_chainlet import stub", "import logging"}
    stub_args = []
    for name, dep in chainlet_descriptor.dependencies.items():
        # `dep.name` is the class name, while `name` is the argument name.
        stub_args.append(f"{name}=stub.factory({dep.name}, self._context)")

    if chainlet_descriptor.has_context:
        if stub_args:
            init_args = f"{', '.join(stub_args)}, context=self._context"
        else:
            init_args = "context=self._context"
    else:
        init_args = ", ".join(stub_args)

    user_chainlet_ref = _gen_chainlet_import_and_ref(chainlet_descriptor)
    imports.update(user_chainlet_ref.imports)
    body = _indent(
        "\n".join(
            [f"logging.info(f'Loading Chainlet `{chainlet_descriptor.name}`.')"]
            + [f"self._chainlet = {user_chainlet_ref.src}({init_args})"]
        )
    )
    src = "\n".join(["def load(self) -> None:", body])
    return _Source(src=src, imports=imports)


def _gen_predict_src(chainlet_descriptor: definitions.ChainletAPIDescriptor) -> _Source:
    """Generates AST for the `predict` method of the truss model."""
    imports: set[str] = {
        "from truss_chains.remote_chainlet import stub",
        "from truss_chains.remote_chainlet import utils",
    }
    parts: list[str] = []
    def_str = "async def" if chainlet_descriptor.endpoint.is_async else "def"
    input_model_name = _get_input_model_name(chainlet_descriptor.name)
    if chainlet_descriptor.endpoint.is_streaming:
        streaming_src = _gen_streaming_type_import_and_ref(
            chainlet_descriptor.endpoint.streaming_type
        )
        imports.update(streaming_src.imports)
        output_type_name = streaming_src.src
    else:
        output_type_name = _get_output_model_name(chainlet_descriptor.name)

    imports.add("import starlette.requests")
    parts.append(
        f"{def_str} predict(self, inputs: {input_model_name}, "
        f"request: starlette.requests.Request) -> {output_type_name}:"
    )
    # Add error handling context manager:
    parts.append(
        _indent("with stub.trace_parent(request), utils.exception_to_http_error():")
    )
    # Invoke Chainlet.
    if (
        chainlet_descriptor.endpoint.is_async
        and not chainlet_descriptor.endpoint.is_streaming
    ):
        maybe_await = "await "
    else:
        maybe_await = ""
    run_remote = chainlet_descriptor.endpoint.name
    # See docs of `pydantic_set_field_dict` for why this is needed.
    args = "**utils.pydantic_set_field_dict(inputs)"
    parts.append(
        _indent(f"result = {maybe_await}self._chainlet.{run_remote}({args})", 2)
    )
    if chainlet_descriptor.endpoint.is_streaming:
        # Streaming returns raw iterator, no pydantic model.
        # This needs to be nested inside the `trace_parent` context!
        parts.append(_indent("async for chunk in result:", 2))
        parts.append(_indent("yield chunk", 3))
    else:
        result_pydantic = f"{output_type_name}(result)"
        parts.append(_indent(f"return {result_pydantic}"))
    return _Source(src="\n".join(parts), imports=imports)


def _gen_truss_chainlet_model(
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
) -> _Source:
    skeleton_tree = libcst.parse_module(_MODEL_SKELETON_FILE.read_text())
    imports: set[str] = set(
        libcst.Module(body=[node]).code
        for node in skeleton_tree.body
        if isinstance(node, libcst.SimpleStatementLine)
        and any(
            isinstance(stmt, libcst.Import) or isinstance(stmt, libcst.ImportFrom)
            for stmt in node.body
        )
    )
    class_definition: libcst.ClassDef = utils.expect_one(
        node
        for node in skeleton_tree.body
        if isinstance(node, libcst.ClassDef) and node.name.value == _MODEL_CLS_NAME
    )

    load_src = _gen_load_src(chainlet_descriptor)
    imports.update(load_src.imports)
    predict_src = _gen_predict_src(chainlet_descriptor)
    imports.update(predict_src.imports)

    new_body: list[Any] = list(class_definition.body.body) + [
        libcst.parse_statement(load_src.src),
        libcst.parse_statement(predict_src.src),
    ]

    user_chainlet_ref = _gen_chainlet_import_and_ref(chainlet_descriptor)
    imports.update(user_chainlet_ref.imports)

    new_block = libcst.IndentedBlock(body=new_body)
    class_definition = class_definition.with_changes(body=new_block)
    class_definition = class_definition.visit(  # type: ignore[assignment]
        _SpecifyChainletTypeAnnotation(user_chainlet_ref.src)
    )
    model_class_src = libcst.Module(body=[class_definition]).code
    return _Source(src=model_class_src, imports=imports)


def _gen_truss_chainlet_file(
    chainlet_dir: pathlib.Path,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    dependencies: Iterable[definitions.ChainletAPIDescriptor],
) -> pathlib.Path:
    """Generates code that wraps a Chainlet as a truss-compatible model."""
    file_path = chainlet_dir / truss_config.DEFAULT_MODEL_MODULE_DIR / _MODEL_FILENAME
    file_path.parent.mkdir(parents=True, exist_ok=True)
    (chainlet_dir / truss_config.DEFAULT_MODEL_MODULE_DIR / "__init__.py").touch()
    imports: set[str] = set()
    src_parts: list[str] = []

    if maybe_stub_src := _gen_stub_src_for_deps(dependencies):
        _update_src(maybe_stub_src, src_parts, imports)

    input_src = _gen_truss_input_pydantic(chainlet_descriptor)
    _update_src(input_src, src_parts, imports)
    if not chainlet_descriptor.endpoint.is_streaming:
        output_src = _gen_truss_output_pydantic(chainlet_descriptor)
        _update_src(output_src, src_parts, imports)
    model_src = _gen_truss_chainlet_model(chainlet_descriptor)
    _update_src(model_src, src_parts, imports)

    imports_str = "\n".join(imports)
    src_str = "\n".join(src_parts)
    file_path.write_text(f"{imports_str}\n{src_str}")
    _format_python_file(file_path)
    return file_path


# Truss Gen ############################################################################


def _make_requirements(image: definitions.DockerImage) -> list[str]:
    """Merges file- and list-based requirements and adds truss git if not present."""
    pip_requirements: set[str] = set()
    if image.pip_requirements_file:
        pip_requirements.update(
            req
            for req in pathlib.Path(image.pip_requirements_file.abs_path)
            .read_text()
            .splitlines()
            if not req.strip().startswith("#")
        )
    pip_requirements.update(image.pip_requirements)

    truss_pypy = next(
        (req for req in pip_requirements if _TRUSS_PIP_PATTERN.match(req)), None
    )

    truss_git = next((req for req in pip_requirements if _TRUSS_GIT in req), None)

    if truss_git:
        logging.warning(
            "The chainlet contains a truss version from github as a pip_requirement:\n"
            f"\t{truss_git}\n"
            "This could result in inconsistencies between the deploying client and the "
            "deployed chainlet. This is not recommended for production chains."
        )
    if truss_pypy:
        logging.warning(
            "The chainlet contains a pinned truss version as a pip_requirement:\n"
            f"\t{truss_pypy}\n"
            "This could result in inconsistencies between the deploying client and the "
            "deployed chainlet. This is not recommended for production chains. If "
            "`truss` is not manually added as a requirement, the same version as "
            "locally installed will be automatically added and ensure compatibility."
        )

    if not (truss_git or truss_pypy):
        truss_pip = f"truss=={truss.version()}"
        logging.debug(
            f"Truss not found in pip requirements, auto-adding: `{truss_pip}`."
        )
        pip_requirements.add(truss_pip)

    return sorted(pip_requirements)


def _inplace_fill_base_image(
    image: definitions.DockerImage, mutable_truss_config: truss_config.TrussConfig
) -> None:
    if isinstance(image.base_image, definitions.BasetenImage):
        mutable_truss_config.python_version = image.base_image.value
    elif isinstance(image.base_image, definitions.CustomImage):
        mutable_truss_config.base_image = truss_config.BaseImage(
            image=image.base_image.image, docker_auth=image.base_image.docker_auth
        )
        if image.base_image.python_executable_path:
            mutable_truss_config.base_image.python_executable_path = (
                image.base_image.python_executable_path
            )
    elif isinstance(image.base_image, str):  # This options is deprecated.
        raise NotImplementedError(
            "Specifying docker base image as string is deprecated"
        )


def _make_truss_config(
    chainlet_dir: pathlib.Path,
    chains_config: definitions.RemoteConfig,
    chainlet_to_service: Mapping[str, definitions.ServiceDescriptor],
    model_name: str,
    use_local_chains_src: bool,
) -> truss_config.TrussConfig:
    """Generate a truss config for a Chainlet."""
    config = truss_config.TrussConfig()
    config.model_name = model_name
    config.model_class_filename = _MODEL_FILENAME
    config.model_class_name = _MODEL_CLS_NAME
    config.runtime.enable_tracing_data = chains_config.options.enable_b10_tracing
    config.environment_variables = dict(chains_config.options.env_variables)
    # Compute.
    compute = chains_config.get_compute_spec()
    config.resources.cpu = str(compute.cpu_count)
    config.resources.memory = str(compute.memory)
    config.resources.accelerator = compute.accelerator
    config.resources.use_gpu = bool(compute.accelerator.count)
    config.runtime.predict_concurrency = compute.predict_concurrency
    # Image.
    _inplace_fill_base_image(chains_config.docker_image, config)
    pip_requirements = _make_requirements(chains_config.docker_image)
    # TODO: `pip_requirements` will add server requirements which give version
    #  conflicts. Check if that's still the case after relaxing versions.
    # config.requirements = pip_requirements
    pip_requirements_file_path = chainlet_dir / _REQUIREMENTS_FILENAME
    pip_requirements_file_path.write_text("\n".join(pip_requirements))
    # Absolute paths don't work with remote build.
    config.requirements_file = _REQUIREMENTS_FILENAME
    config.system_packages = chains_config.docker_image.apt_requirements
    if chains_config.docker_image.external_package_dirs:
        for ext_dir in chains_config.docker_image.external_package_dirs:
            config.external_package_dirs.append(ext_dir.abs_path)
    config.use_local_chains_src = use_local_chains_src
    # Assets.
    assets = chains_config.get_asset_spec()
    config.secrets = assets.secrets
    if definitions.BASETEN_API_SECRET_NAME not in config.secrets:
        config.secrets[definitions.BASETEN_API_SECRET_NAME] = definitions.SECRET_DUMMY
    else:
        logging.info(
            f"Chains automatically add {definitions.BASETEN_API_SECRET_NAME} "
            "to secrets - no need to manually add it."
        )
    config.model_cache.models = assets.cached
    config.external_data = truss_config.ExternalData(items=assets.external_data)
    # Metadata.
    chains_metadata: definitions.TrussMetadata = definitions.TrussMetadata(
        chainlet_to_service=chainlet_to_service
    )
    config.model_metadata[definitions.TRUSS_CONFIG_CHAINS_KEY] = (
        chains_metadata.model_dump()
    )
    config.write_to_yaml_file(
        chainlet_dir / serving_image_builder.CONFIG_FILE, verbose=True
    )
    return config


def gen_truss_chainlet(
    chain_root: pathlib.Path,
    gen_root: pathlib.Path,
    chain_name: str,
    chainlet_descriptor: definitions.ChainletAPIDescriptor,
    model_name: str,
    use_local_chains_src: bool,
) -> pathlib.Path:
    # Filter needed services and customize options.
    dep_services = {}
    for dep in chainlet_descriptor.dependencies.values():
        dep_services[dep.name] = definitions.ServiceDescriptor(
            name=dep.name,
            display_name=dep.display_name,
            options=dep.options,
        )
    chainlet_dir = _make_chainlet_dir(chain_name, chainlet_descriptor, gen_root)
    logging.info(
        f"Code generation for Chainlet `{chainlet_descriptor.name}` "
        f"in `{chainlet_dir}`."
    )
    _make_truss_config(
        chainlet_dir,
        chainlet_descriptor.chainlet_cls.remote_config,
        dep_services,
        model_name,
        use_local_chains_src,
    )
    # TODO This assumes all imports are absolute w.r.t chain root (or site-packages).
    truss_path.copy_tree_path(
        chain_root, chainlet_dir / truss_config.DEFAULT_BUNDLED_PACKAGES_DIR
    )
    for file in chain_root.glob("*.py"):
        if "-" in file.name:
            raise definitions.ChainsUsageError(
                f"Python file `{file}` contains `-`, use `_` instead."
            )
        if file.name == _MODEL_FILENAME:
            raise definitions.ChainsUsageError(
                f"Python file name `{_MODEL_FILENAME}` is reserved and cannot be used."
            )
    chainlet_file = _gen_truss_chainlet_file(
        chainlet_dir,
        chainlet_descriptor,
        framework.get_dependencies(chainlet_descriptor),
    )
    remote_config = chainlet_descriptor.chainlet_cls.remote_config
    if remote_config.docker_image.data_dir:
        data_dir = chainlet_dir / truss_config.DEFAULT_DATA_DIRECTORY
        data_dir.mkdir(parents=True, exist_ok=True)
        user_data_dir = remote_config.docker_image.data_dir.abs_path
        shutil.copytree(user_data_dir, data_dir, dirs_exist_ok=True)

    # Copy model file s.t. during debugging imports can are properly resolved.
    shutil.copy(
        chainlet_file,
        chainlet_file.parent.parent
        / truss_config.DEFAULT_BUNDLED_PACKAGES_DIR
        / "_model_dbg.py",
    )
    return chainlet_dir



from briton.spec_dec_truss_model import Model as SpecDecModel
from briton.trtllm_config import (
    TRTLLMConfiguration,
)
from briton.truss_model import Model

# TODO(pankaj) Define an ABC base class for this. That baseclass should live in
# a new, smaller truss sub-library, perhaps called `truss-runtime`` for inclusion
# in Truss runtime. Once we have that sub-library, we should define the Extension
# base class there and derive Extension class below from it.
#
# That base class would look like:
# class TrussExtension(ABC):
#     @abstracemethod
#     def model_override(self):
#         pass

#     @abstractmethod
#     def model_args(self) -> dict:
#         pass


#     @abstractmethod
#     def load(self) -> dict:
#         pass
class Extension:
    """
    trt_llm truss extension.

    Provides model_args to supply to model class, which contain the trtllm
    engine that corresponds to provided config.

    This extension also provides a full replacement of the model class, which is
    to be used if user doesn't supply it. This may be desired behavior in many
    cases where users want to just go by config and don't want to do any pre or
    post-processing.
    """

    def __init__(self, *args, **kwargs):
        self._config = kwargs["config"]
        trt_llm_config = self._config.get("trt_llm")
        config = TRTLLMConfiguration(**trt_llm_config)
        if config.build.speculator:
            self._model = SpecDecModel(*args, **kwargs)
        else:
            self._model = Model(*args, **kwargs)

    def model_override(self):
        """Return a model object.

        This is used if model.py is omitted, which is allowed when using trt_llm.
        """
        return self._model

    def model_args(self) -> dict:
        """Return args to supply as input to Model class' __init__ method.

        Model class can use this to invoke the trt_llm engine.

        Returned engine is a typical Truss model class that provides a predict
        function. The predict function is async and returns an async generator.
        """
        return {"engine": self._model}

    def load(self):
        self._model.load()



import os

from fastapi import FastAPI

app = FastAPI()


@app.get("/health")
async def health():
    return {"message": "OK"}


@app.post("/predict")
async def root():
    return {
        "message": "Hello World",
        "is_env_var_passed": os.environ.get("HF_TOKEN") is not None,
        "is_secret_mounted": os.path.exists("/secrets/hf_access_token"),
    }



import asyncio
import dataclasses
import enum
import struct
import sys
from collections.abc import AsyncIterator
from typing import Generic, Optional, Protocol, Type, TypeVar, overload

import pydantic

_TAG_SIZE = 5  # uint8 + uint32.

_T = TypeVar("_T")

if sys.version_info < (3, 10):

    async def anext(iterable: AsyncIterator[_T]) -> _T:
        return await iterable.__anext__()


# Note on the (verbose) typing in this module: we want exact typing of the reader and
# writer helpers, while also allowing flexibility to users to leave out header/footer
# if not needed.
# Putting both a constraint on the header/footer types to be pydantic
# models, but also letting them be optional is not well-supported by typing tools,
# (missing feature is using type variables a constraints on other type variables).
#
# A functional, yet verbose workaround that gives correct variadic type inference,
# is using intermediate type variables `HeaderT` <-> `HeaderTT` and in conjunction with
# mapping out all usage combinations with overloads (the overloads essentially allow
# "conditional" binding of type vars). These overloads also allow to use granular
# reader/writer sub-classes conditionally, that have the read/write methods only for the
# data types configured, and implemented DRY with mixin classes.
ItemT = TypeVar("ItemT", bound=pydantic.BaseModel)
HeaderT = TypeVar("HeaderT", bound=pydantic.BaseModel)
FooterT = TypeVar("FooterT", bound=pydantic.BaseModel)

# Since header/footer could also be `None`, we need an extra type variable that
# can assume either `Type[HeaderT]` or `None` - `Type[None]` causes issues.
HeaderTT = TypeVar("HeaderTT")
FooterTT = TypeVar("FooterTT")


@dataclasses.dataclass
class StreamTypes(Generic[ItemT, HeaderTT, FooterTT]):
    item_type: Type[ItemT]
    header_type: HeaderTT  # Is either `Type[HeaderT]` or `None`.
    footer_type: FooterTT  # Is either `Type[FooterT]` or `None`.


@overload
def stream_types(
    item_type: Type[ItemT],
    *,
    header_type: Type[HeaderT],
    footer_type: Type[FooterT],
) -> StreamTypes[ItemT, HeaderT, FooterT]: ...


@overload
def stream_types(
    item_type: Type[ItemT],
    *,
    header_type: Type[HeaderT],
) -> StreamTypes[ItemT, HeaderT, None]: ...


@overload
def stream_types(
    item_type: Type[ItemT],
    *,
    footer_type: Type[FooterT],
) -> StreamTypes[ItemT, None, FooterT]: ...


@overload
def stream_types(item_type: Type[ItemT]) -> StreamTypes[ItemT, None, None]: ...


def stream_types(
    item_type: Type[ItemT],
    *,
    header_type: Optional[Type[HeaderT]] = None,
    footer_type: Optional[Type[FooterT]] = None,
) -> StreamTypes:
    """Creates a bundle of item type and potentially header/footer types,
    each as pydantic model."""
    # This indirection for creating `StreamTypes` is needed to get generic typing.
    return StreamTypes(item_type, header_type, footer_type)


# Reading ##############################################################################


class _Delimiter(enum.IntEnum):
    NOT_SET = enum.auto()
    HEADER = enum.auto()
    ITEM = enum.auto()
    FOOTER = enum.auto()
    END = enum.auto()


class _Streamer(Generic[ItemT, HeaderTT, FooterTT]):
    _stream_types: StreamTypes[ItemT, HeaderTT, FooterTT]

    def __init__(self, types: StreamTypes[ItemT, HeaderTT, FooterTT]) -> None:
        self._stream_types = types


# Reading ##############################################################################


class _ByteReader:
    """Helper to provide `readexactly` API for an async bytes iterator."""

    def __init__(self, source: AsyncIterator[bytes]) -> None:
        self._source = source
        self._buffer = bytearray()

    async def readexactly(self, num_bytes: int) -> bytes:
        while len(self._buffer) < num_bytes:
            try:
                chunk = await anext(self._source)
            except StopAsyncIteration:
                break
            self._buffer.extend(chunk)

        if len(self._buffer) < num_bytes:
            if len(self._buffer) == 0:
                raise EOFError()
            raise asyncio.IncompleteReadError(self._buffer, num_bytes)

        result = bytes(self._buffer[:num_bytes])
        del self._buffer[:num_bytes]
        return result


class _StreamReaderProtocol(Protocol[ItemT, HeaderTT, FooterTT]):
    _stream_types: StreamTypes[ItemT, HeaderTT, FooterTT]
    _footer_data: Optional[bytes]

    async def _read(self) -> tuple[_Delimiter, bytes]: ...


class _StreamReader(_Streamer[ItemT, HeaderTT, FooterTT]):
    _stream: _ByteReader
    _footer_data: Optional[bytes]

    def __init__(
        self,
        types: StreamTypes[ItemT, HeaderTT, FooterTT],
        stream: AsyncIterator[bytes],
    ) -> None:
        super().__init__(types)
        self._stream = _ByteReader(stream)
        self._footer_data = None

    @staticmethod
    def _unpack_tag(tag: bytes) -> tuple[_Delimiter, int]:
        enum_value, length = struct.unpack(">BI", tag)
        return _Delimiter(enum_value), length

    async def _read(self) -> tuple[_Delimiter, bytes]:
        try:
            tag = await self._stream.readexactly(_TAG_SIZE)
        # It's ok to read nothing (end of stream), but unexpected to read partial.
        except asyncio.IncompleteReadError:
            raise
        except EOFError:
            return _Delimiter.END, b""

        delimiter, length = self._unpack_tag(tag)
        if not length:
            return delimiter, b""
        data_bytes = await self._stream.readexactly(length)
        return delimiter, data_bytes

    async def read_items(self) -> AsyncIterator[ItemT]:
        delimiter, data_bytes = await self._read()
        if delimiter == _Delimiter.HEADER:
            raise ValueError(
                "Called `read_items`, but there the stream contains header data, which "
                "is not consumed. Call `read_header` first or remove sending a header."
            )
        if delimiter in (_Delimiter.FOOTER, _Delimiter.END):  # In case of 0 items.
            self._footer_data = data_bytes
            return

        assert delimiter == _Delimiter.ITEM
        while True:
            yield self._stream_types.item_type.model_validate_json(data_bytes)
            # We don't know if the next data is another item, footer or the end.
            delimiter, data_bytes = await self._read()
            if delimiter == _Delimiter.END:
                return
            if delimiter == _Delimiter.FOOTER:
                self._footer_data = data_bytes
                return


class _HeaderReadMixin(_Streamer[ItemT, HeaderT, FooterTT]):
    async def read_header(
        self: _StreamReaderProtocol[ItemT, HeaderT, FooterTT],
    ) -> HeaderT:
        delimiter, data_bytes = await self._read()
        if delimiter != _Delimiter.HEADER:
            raise ValueError("Stream does not contain header.")
        return self._stream_types.header_type.model_validate_json(data_bytes)


class _FooterReadMixin(_Streamer[ItemT, HeaderTT, FooterT]):
    _footer_data: Optional[bytes]

    async def read_footer(
        self: _StreamReaderProtocol[ItemT, HeaderTT, FooterT],
    ) -> FooterT:
        if self._footer_data is None:
            delimiter, data_bytes = await self._read()
            if delimiter != _Delimiter.FOOTER:
                raise ValueError("Stream does not contain footer.")
            self._footer_data = data_bytes

        footer = self._stream_types.footer_type.model_validate_json(self._footer_data)
        self._footer_data = None
        return footer


class StreamReaderWithHeader(
    _StreamReader[ItemT, HeaderT, FooterTT], _HeaderReadMixin[ItemT, HeaderT, FooterTT]
): ...


class StreamReaderWithFooter(
    _StreamReader[ItemT, HeaderTT, FooterT], _FooterReadMixin[ItemT, HeaderTT, FooterT]
): ...


class StreamReaderFull(
    _StreamReader[ItemT, HeaderT, FooterT],
    _HeaderReadMixin[ItemT, HeaderT, FooterT],
    _FooterReadMixin[ItemT, HeaderT, FooterT],
): ...


@overload
def stream_reader(
    types: StreamTypes[ItemT, None, None],
    stream: AsyncIterator[bytes],
) -> _StreamReader[ItemT, None, None]: ...


@overload
def stream_reader(
    types: StreamTypes[ItemT, HeaderT, None],
    stream: AsyncIterator[bytes],
) -> StreamReaderWithHeader[ItemT, HeaderT, None]: ...


@overload
def stream_reader(
    types: StreamTypes[ItemT, None, FooterT],
    stream: AsyncIterator[bytes],
) -> StreamReaderWithFooter[ItemT, None, FooterT]: ...


@overload
def stream_reader(
    types: StreamTypes[ItemT, HeaderT, FooterT],
    stream: AsyncIterator[bytes],
) -> StreamReaderFull[ItemT, HeaderT, FooterT]: ...


def stream_reader(
    types: StreamTypes[ItemT, HeaderTT, FooterTT],
    stream: AsyncIterator[bytes],
) -> _StreamReader:
    if types.header_type is None and types.footer_type is None:
        return _StreamReader(types, stream)
    if types.header_type is None:
        return StreamReaderWithFooter(types, stream)
    if types.footer_type is None:
        return StreamReaderWithHeader(types, stream)

    return StreamReaderFull(types, stream)


# Writing ##############################################################################


class _StreamWriterProtocol(Protocol[ItemT, HeaderTT, FooterTT]):
    _stream_types: StreamTypes[ItemT, HeaderTT, FooterTT]
    _last_sent: _Delimiter

    def _serialize(self, obj: pydantic.BaseModel, delimiter: _Delimiter) -> bytes: ...


class _StreamWriter(_Streamer[ItemT, HeaderTT, FooterTT]):
    def __init__(self, types: StreamTypes[ItemT, HeaderTT, FooterTT]) -> None:
        super().__init__(types)
        self._last_sent = _Delimiter.NOT_SET
        self._stream_types = types

    @staticmethod
    def _pack_tag(delimiter: _Delimiter, length: int) -> bytes:
        return struct.pack(">BI", delimiter.value, length)

    def _serialize(self, obj: pydantic.BaseModel, delimiter: _Delimiter) -> bytes:
        data_bytes = obj.model_dump_json().encode()
        data = bytearray(self._pack_tag(delimiter, len(data_bytes)))
        data.extend(data_bytes)
        # Starlette cannot handle byte array, but view works..
        return memoryview(data)

    def yield_item(self, item: ItemT) -> bytes:
        if self._last_sent in (_Delimiter.FOOTER, _Delimiter.END):
            raise ValueError("Cannot yield item after sending footer / closing stream.")
        self._last_sent = _Delimiter.ITEM
        return self._serialize(item, _Delimiter.ITEM)


class _HeaderWriteMixin(_Streamer[ItemT, HeaderT, FooterTT]):
    def yield_header(
        self: _StreamWriterProtocol[ItemT, HeaderT, FooterTT], header: HeaderT
    ) -> bytes:
        if self._last_sent != _Delimiter.NOT_SET:
            raise ValueError("Cannot yield header after other data has been sent.")
        self._last_sent = _Delimiter.HEADER
        return self._serialize(header, _Delimiter.HEADER)


class _FooterWriteMixin(_Streamer[ItemT, HeaderTT, FooterT]):
    def yield_footer(
        self: _StreamWriterProtocol[ItemT, HeaderTT, FooterT], footer: FooterT
    ) -> bytes:
        if self._last_sent == _Delimiter.END:
            raise ValueError("Cannot yield footer after closing stream.")
        self._last_sent = _Delimiter.FOOTER
        return self._serialize(footer, _Delimiter.FOOTER)


class StreamWriterWithHeader(
    _StreamWriter[ItemT, HeaderT, FooterTT], _HeaderWriteMixin[ItemT, HeaderT, FooterTT]
): ...


class StreamWriterWithFooter(
    _StreamWriter[ItemT, HeaderTT, FooterT], _FooterWriteMixin[ItemT, HeaderTT, FooterT]
): ...


class StreamWriterFull(
    _StreamWriter[ItemT, HeaderT, FooterT],
    _HeaderWriteMixin[ItemT, HeaderT, FooterT],
    _FooterWriteMixin[ItemT, HeaderT, FooterT],
): ...


@overload
def stream_writer(
    types: StreamTypes[ItemT, None, None],
) -> _StreamWriter[ItemT, None, None]: ...


@overload
def stream_writer(
    types: StreamTypes[ItemT, HeaderT, None],
) -> StreamWriterWithHeader[ItemT, HeaderT, None]: ...


@overload
def stream_writer(
    types: StreamTypes[ItemT, None, FooterT],
) -> StreamWriterWithFooter[ItemT, None, FooterT]: ...


@overload
def stream_writer(
    types: StreamTypes[ItemT, HeaderT, FooterT],
) -> StreamWriterFull[ItemT, HeaderT, FooterT]: ...


def stream_writer(
    types: StreamTypes[ItemT, HeaderTT, FooterTT],
) -> _StreamWriter:
    if types.header_type is None and types.footer_type is None:
        return _StreamWriter(types)
    if types.header_type is None:
        return StreamWriterWithFooter(types)
    if types.footer_type is None:
        return StreamWriterWithHeader(types)

    return StreamWriterFull(types)



import functools
import pathlib
from typing import TYPE_CHECKING, ContextManager, Mapping, Optional, Type, Union

from truss_chains import definitions, framework
from truss_chains.deployment import deployment_client

if TYPE_CHECKING:
    from rich import progress


def depends_context() -> definitions.DeploymentContext:
    """Sets a "symbolic marker" for injecting a context object at runtime.

    Refer to `the docs <https://docs.baseten.co/chains/getting-started>`_ and this
    `example chainlet <https://github.com/basetenlabs/truss/blob/main/truss-chains/truss_chains/example_chainlet.py>`_
    for more guidance on the ``__init__``-signature of chainlets.

    Warning:
        Despite the type annotation, this does *not* immediately provide a
        context instance. Only when deploying remotely or using ``run_local`` a
        context instance is provided.

    Returns:
        A "symbolic marker" to be used as a default argument in a chainlet's
        initializer.


    """
    # The type error is silenced to because chains framework will at runtime inject
    # a corresponding instance. Nonetheless, we want to use a type annotation here,
    # to facilitate type inference, code-completion and type checking within the code
    # of chainlets that depend on the other chainlet.
    return framework.ContextDependencyMarker()  # type: ignore


def depends(
    chainlet_cls: Type[framework.ChainletT],
    retries: int = 1,
    timeout_sec: int = definitions.DEFAULT_TIMEOUT_SEC,
    use_binary: bool = False,
) -> framework.ChainletT:
    """Sets a "symbolic marker" to indicate to the framework that a chainlet is a
    dependency of another chainlet. The return value of ``depends`` is intended to be
    used as a default argument in a chainlet's ``__init__``-method.
    When deploying a chain remotely, a corresponding stub to the remote is injected in
    its place. In ``run_local`` mode an instance of a local chainlet is injected.

    Refer to `the docs <https://docs.baseten.co/chains/getting-started>`_ and this
    `example chainlet <https://github.com/basetenlabs/truss/blob/main/truss-chains/truss_chains/example_chainlet.py>`_
    for more guidance on how make one chainlet depend on another chainlet.

    Warning:
        Despite the type annotation, this does *not* immediately provide a
        chainlet instance. Only when deploying remotely or using ``run_local`` a
        chainlet instance is provided.


    Args:
        chainlet_cls: The chainlet class of the dependency.
        retries: The number of times to retry the remote chainlet in case of failures
          (e.g. due to transient network issues). For streaming, retries are only made
          if the request fails before streaming any results back. Failures mid-stream
          not retried.
        timeout_sec: Timeout for the HTTP request to this chainlet.
        use_binary: whether to send data data in binary format. This can give a parsing
         speedup and message size reduction (~25%) for numpy arrays. Use
         ``NumpyArrayField`` as a field type on pydantic models for integration and set
         this option to ``True``. For simple text data, there is no significant benefit.

    Returns:
        A "symbolic marker" to be used as a default argument in a chainlet's
        initializer.
    """
    options = definitions.RPCOptions(
        retries=retries, timeout_sec=timeout_sec, use_binary=use_binary
    )
    # The type error is silenced to because chains framework will at runtime inject
    # a corresponding instance. Nonetheless, we want to use a type annotation here,
    # to facilitate type inference, code-completion and type checking within the code
    # of chainlets that depend on the other chainlet.
    return framework.ChainletDependencyMarker(chainlet_cls, options)  # type: ignore


class ChainletBase(definitions.ABCChainlet):
    """Base class for all chainlets.

    Inheriting from this class adds validations to make sure subclasses adhere to the
    chainlet pattern and facilitates remote chainlet deployment.

    Refer to `the docs <https://docs.baseten.co/chains/getting-started>`_ and this
    `example chainlet <https://github.com/basetenlabs/truss/blob/main/truss-chains/truss_chains/example_chainlet.py>`_
    for more guidance on how to create subclasses.
    """

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        framework.validate_and_register_class(cls)  # Errors are collected, not raised!
        # For default init (from `object`) we don't need to check anything.
        if cls.has_custom_init():
            original_init = cls.__init__

            @functools.wraps(original_init)
            def __init_with_arg_check__(self, *args, **kwargs):
                if args:
                    raise definitions.ChainsRuntimeError("Only kwargs are allowed.")
                framework.ensure_args_are_injected(cls, original_init, kwargs)
                original_init(self, *args, **kwargs)

            cls.__init__ = __init_with_arg_check__  # type: ignore[method-assign]


def mark_entrypoint(cls: Type[framework.ChainletT]) -> Type[framework.ChainletT]:
    """Decorator to mark a chainlet as the entrypoint of a chain.

    This decorator can be applied to *one* chainlet in a source file and then the
    CLI push command simplifies because only the file, but not the chainlet class
    in the file, needs to be specified.

    Example usage::

        import truss_chains as chains

        @chains.mark_entrypoint
        class MyChainlet(ChainletBase):
            ...
    """
    return framework.entrypoint(cls)


def push(
    entrypoint: Type[definitions.ABCChainlet],
    chain_name: str,
    publish: bool = True,
    promote: bool = True,
    only_generate_trusses: bool = False,
    remote: str = "baseten",
    environment: Optional[str] = None,
    progress_bar: Optional[Type["progress.Progress"]] = None,
) -> deployment_client.BasetenChainService:
    """
    Deploys a chain remotely (with all dependent chainlets).

    Args:
        entrypoint: The chainlet class that serves as the entrypoint to the chain.
        chain_name: The name of the chain.
        publish: Whether to publish the chain as a published deployment (it is a
          draft deployment otherwise)
        promote: Whether to promote the chain to be the production deployment (this
          implies publishing as well).
        only_generate_trusses: Used for debugging purposes. If set to True, only the
          the underlying truss models for the chainlets are generated in
          ``/tmp/.chains_generated``.
        remote: name of a remote config in `.trussrc`. If not provided, it will be
          inquired.
        environment: The name of an environment to promote deployment into.
        progress_bar: Optional `rich.progress.Progress` if output is desired.

    Returns:
        A chain service handle to the deployed chain.

    """
    options = definitions.PushOptionsBaseten.create(
        chain_name=chain_name,
        publish=publish,
        promote=promote,
        only_generate_trusses=only_generate_trusses,
        remote=remote,
        environment=environment,
    )
    service = deployment_client.push(entrypoint, options, progress_bar=progress_bar)
    assert isinstance(
        service, deployment_client.BasetenChainService
    )  # Per options above.
    return service


def run_local(
    secrets: Optional[Mapping[str, str]] = None,
    data_dir: Optional[Union[pathlib.Path, str]] = None,
    chainlet_to_service: Optional[
        Mapping[str, definitions.DeployedServiceDescriptor]
    ] = None,
) -> ContextManager[None]:
    """Context manager local debug execution of a chain.

    The arguments only need to be provided if the chainlets explicitly access any the
    corresponding fields of ``DeploymentContext``.

    Args:
        secrets: A dict of secrets keys and values to provide to the chainlets.
        data_dir: Path to a directory with data files.
        chainlet_to_service: A dict of chainlet names to service descriptors.

    Example usage (as trailing main section in a chain file)::

        import os
        import truss_chains as chains


        class HelloWorld(chains.ChainletBase):
            ...


        if __name__ == "__main__":
            with chains.run_local(
                secrets={"some_token": os.environ["SOME_TOKEN"]},
                chainlet_to_service={
                    "SomeChainlet": chains.DeployedServiceDescriptor(
                        name="SomeChainlet",
                        display_name="SomeChainlet",
                        predict_url="https://...",
                        options=chains.RPCOptions(),
                    )
                },
            ):
                hello_world_chain = HelloWorld()
                result = hello_world_chain.run_remote(max_value=5)

            print(result)


    Refer to the `local debugging guide <https://docs.baseten.co/chains/guide#test-a-chain-locally>`_
    for more details.
    """
    data_dir = pathlib.Path(data_dir) if data_dir else None
    return framework.run_local(secrets or {}, data_dir, chainlet_to_service or {})



import logging
import os
import sys
from pathlib import Path
from typing import List, Optional

import yaml

from truss.base.constants import CONFIG_FILE, TEMPLATES_DIR
from truss.base.truss_config import (
    Build,
    TrussConfig,
    map_local_to_supported_python_version,
)
from truss.truss_handle.truss_handle import TrussHandle
from truss.util.notebook import is_notebook_or_ipython
from truss.util.path import build_truss_target_directory, copy_tree_path

logger: logging.Logger = logging.getLogger(__name__)

if is_notebook_or_ipython():
    logger.setLevel(logging.INFO)
    logger.addHandler(logging.StreamHandler(sys.stdout))


def _populate_target_directory(
    config: TrussConfig,
    target_directory_path: Optional[str] = None,
    template: str = "custom",
    populate_dirs: bool = True,
) -> Path:
    target_directory_path_typed = None
    if target_directory_path is None:
        target_directory_path_typed = build_truss_target_directory(template)
    else:
        target_directory_path_typed = Path(target_directory_path)
        target_directory_path_typed.mkdir(parents=True, exist_ok=True)

    if populate_dirs:
        # Create data dir
        (target_directory_path_typed / config.data_dir).mkdir()

        # Create bundled packages dir
        # TODO: Drop by default
        (target_directory_path_typed / config.bundled_packages_dir).mkdir()

        # Create model module dir
        model_dir = target_directory_path_typed / config.model_module_dir
        template_path = TEMPLATES_DIR / template
        copy_tree_path(template_path / "model", model_dir)

    # Write config
    with (target_directory_path_typed / CONFIG_FILE).open("w") as config_file:
        yaml.dump(config.to_dict(verbose=False), config_file)

    return target_directory_path_typed


def init(
    target_directory: str,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
    build_config: Optional[Build] = None,
    model_name: Optional[str] = None,
) -> TrussHandle:
    """
    Initialize an empty placeholder Truss. A Truss is a build context designed
    to be built as a container locally or uploaded into a baseten serving
    environment. This placeholder structure can be filled to represent ML
    models.

    Args:
        target_directory: Absolute or relative path of the directory to create
                          Truss in. The directory is created if it doesn't exist.
    """
    config = TrussConfig(
        model_name=model_name,
        python_version=map_local_to_supported_python_version(),
    )

    if build_config:
        config.build = build_config

    target_directory_path = _populate_target_directory(
        config=config,
        target_directory_path=target_directory,
        populate_dirs=True,
    )

    scaf = TrussHandle(target_directory_path)
    _update_truss_props(scaf, data_files, requirements_file, bundled_packages)
    return scaf


def load(truss_directory: str) -> TrussHandle:
    """Get a handle to a Truss. A Truss is a build context designed to be built
    as a container locally or uploaded into a model serving environment.

    Args:
        truss_directory (str): The local directory of an existing Truss
    Returns:
        TrussHandle
    """
    return TrussHandle(Path(truss_directory))


def cleanup() -> None:
    """
    Cleans up .truss directory.
    """
    build_folder_path = Path(Path.home(), ".truss")
    if build_folder_path.exists():
        for obj in build_folder_path.glob("**/*"):
            if (not obj.name == "config.yaml") and (obj.is_file()):
                os.remove(obj)
    return


def _update_truss_props(
    scaf: TrussHandle,
    data_files: Optional[List[str]] = None,
    requirements_file: Optional[str] = None,
    bundled_packages: Optional[List[str]] = None,
) -> None:
    if data_files is not None:
        for data_file in data_files:
            scaf.add_data(data_file)

    if bundled_packages is not None:
        for package in bundled_packages:
            scaf.add_bundled_package(package)

    if requirements_file is not None:
        scaf.update_requirements_from_file(requirements_file)



from pathlib import Path

from truss.base.truss_spec import TrussSpec
from truss.truss_handle.build import init


def test_truss_init(tmp_path):
    dir_name = str(tmp_path)
    init(dir_name)
    spec = TrussSpec(Path(dir_name))
    assert spec.model_module_dir.exists()
    assert spec.data_dir.exists()
    assert spec.truss_dir == tmp_path
    assert spec.config_path.exists()


def test_truss_init_with_data_file_and_requirements_file_and_bundled_packages(
    tmp_path,
):
    dir_path = tmp_path / "truss"
    dir_name = str(dir_path)

    # Init data files
    data_path = tmp_path / "data.txt"
    with data_path.open("w") as data_file:
        data_file.write("test")

    # Init requirements file
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")

    # init bundled packages
    packages_path = tmp_path / "dep_pkg"
    packages_path.mkdir()
    packages_path_file_py = packages_path / "file.py"
    packages_path_init_py = packages_path / "__init__.py"
    pkg_files = [packages_path_init_py, packages_path_file_py]
    for pkg_file in pkg_files:
        with pkg_file.open("w") as fh:
            fh.write("test")

    init(
        dir_name,
        data_files=[str(data_path)],
        requirements_file=str(req_file_path),
        bundled_packages=[str(packages_path)],
    )
    spec = TrussSpec(Path(dir_name))
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert spec.bundled_packages_dir.exists()
    assert (spec.data_dir / "data.txt").exists()
    assert spec.requirements == requirements
    assert (spec.bundled_packages_dir / "dep_pkg" / "__init__.py").exists()
    assert (spec.bundled_packages_dir / "dep_pkg" / "file.py").exists()
